#!/usr/bin/python3

from brownie._config import CONFIG as config
from brownie.convert import Fixed, Wei
from brownie.project import compile_source, run
from brownie.network import accounts, alert, history, rpc, web3
from brownie.network.contract import Contract  # NOQA: F401

__all__ = [
    "accounts",  # accounts is an Accounts singleton
    "alert",
    "history",  # history is a TxHistory singleton
    "network",
    "rpc",  # rpc is a Rpc singleton
    "web3",  # web3 is a Web3 instance
    "project",
    "compile_source",
    "run",
    "Fixed",
    "Wei",
    "config",
]



#!/usr/bin/python3

import ast
import importlib
import sys
from hashlib import sha1
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Optional, Tuple

from brownie.exceptions import ProjectNotFound
from brownie.project.main import Project, check_for_project, get_loaded_projects
from brownie.utils import color

_import_cache: Dict = {}


def run(
    script_path: str,
    method_name: str = "main",
    args: Optional[Tuple] = None,
    kwargs: Optional[Dict] = None,
    project: Any = None,
) -> None:
    """Loads a project script and runs a method in it.

    script_path: path of script to load
    method_name: name of method to run
    args: method args
    kwargs: method kwargs
    project: (deprecated)

    Returns: return value from called method
    """
    if args is None:
        args = tuple()
    if kwargs is None:
        kwargs = {}

    if not get_loaded_projects():
        raise ProjectNotFound("Cannot run a script without an active project")

    script, project = _get_path(script_path)

    # temporarily add project objects to the main namespace, so the script can import them
    brownie: Any = sys.modules["brownie"]
    brownie_dict = brownie.__dict__.copy()
    brownie_all = brownie.__all__.copy()
    brownie.__dict__.update(project)
    brownie.__all__.extend(project.__all__)

    try:
        script = script.absolute().relative_to(project._path)
        module = _import_from_path(script)

        name = module.__name__
        if not hasattr(module, method_name):
            raise AttributeError(f"Module '{name}' has no method '{method_name}'")
        print(
            f"\nRunning '{color['module']}{name}{color}.{color['callable']}{method_name}{color}'..."
        )
        return getattr(module, method_name)(*args, **kwargs)
    finally:
        # cleanup namespace
        brownie.__dict__.clear()
        brownie.__dict__.update(brownie_dict)
        brownie.__all__ = brownie_all


def _get_path(path_str: str) -> Tuple[Path, Project]:
    # Returns path to a python module
    path = Path(path_str).with_suffix(".py")

    if not path.is_absolute():
        if path.parts[0] != "scripts":
            path = Path("scripts").joinpath(path)
        for project in get_loaded_projects():
            if project._path.joinpath(path).exists():
                path = project._path.joinpath(path)
                return path, project
        raise FileNotFoundError(f"Cannot find {path_str}")

    if not path.exists():
        raise FileNotFoundError(f"Cannot find {path_str}")

    try:
        project = next(i for i in get_loaded_projects() if path_str.startswith(i._path.as_posix()))
    except StopIteration:
        raise ProjectNotFound(f"{path_str} is not part of an active project")

    return path, project


def _import_from_path(path: Path) -> ModuleType:
    # Imports a module from the given path
    import_str = ".".join(path.parts[:-1] + (path.stem,))
    if import_str in _import_cache:
        importlib.reload(_import_cache[import_str])
    else:
        _import_cache[import_str] = importlib.import_module(import_str)
    return _import_cache[import_str]


def _get_ast_hash(path: str) -> str:
    # Generates a hash based on the AST of a script.
    with Path(path).open() as fp:
        ast_list = [ast.parse(fp.read(), path)]
    base_path = str(check_for_project(path))
    for obj in [i for i in ast_list[0].body if type(i) in (ast.Import, ast.ImportFrom)]:
        if type(obj) is ast.Import:
            name = obj.names[0].name  # type: ignore
        else:
            name = obj.module  # type: ignore
        try:
            origin = importlib.util.find_spec(name).origin
        except Exception as e:
            raise type(e)(f"in {path} - {e}") from None
        if base_path in origin:
            with open(origin) as fp:
                ast_list.append(ast.parse(fp.read(), origin))
    dump = "\n".join(ast.dump(i) for i in ast_list)
    return sha1(dump.encode()).hexdigest()



#!/usr/bin/python3

import hashlib
import itertools
import json
import re
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple
from urllib.parse import urlparse

from ethpm._utils.ipfs import generate_file_hash
from ethpm.backends.ipfs import InfuraIPFSBackend

from brownie import network
from brownie._config import _get_data_folder
from brownie.convert import to_address
from brownie.exceptions import InvalidManifest
from brownie.network.web3 import web3
from brownie.typing import AccountsType, TransactionReceiptType
from brownie.utils import color

from . import compiler

URI_REGEX = r"""^(?:erc1319://|)([^/:\s]*)(?::[0-9]+|)/([a-z][a-z0-9_-]{0,255})@([^\s:/'";]*)$"""

REGISTRY_ABI = [
    {
        "constant": False,
        "inputs": [
            {"name": "packageName", "type": "string"},
            {"name": "version", "type": "string"},
            {"name": "manifestURI", "type": "string"},
        ],
        "name": "release",
        "outputs": [{"name": "releaseId", "type": "bytes32"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    }
]


def get_manifest(uri: str) -> Dict:

    """
    Fetches an ethPM manifest and processes it for use with Brownie.
    A local copy is also stored if the given URI follows the ERC1319 spec.

    Args:
        uri: URI location of the manifest. Can be IPFS or ERC1319.
    """

    # uri can be a registry uri or a direct link to ipfs
    if not isinstance(uri, str):
        raise TypeError("EthPM manifest uri must be given as a string")

    match = re.match(URI_REGEX, uri)
    if match is None:
        # if a direct link to IPFS was used, we don't save the manifest locally
        manifest = json.loads(_get_uri_contents(uri))
        path = None
    else:
        address, package_name, version = match.groups()
        path = _get_data_folder().joinpath(f"ethpm/{address}/{package_name}/{version}.json")
        try:
            with path.open("r") as fp:
                return json.load(fp)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            pass
        # TODO chain != 1
        web3._mainnet.pm.set_registry(address)
        package = web3._mainnet.pm.get_package(package_name, version)
        manifest = package.manifest
        uri = package.uri

    manifest = process_manifest(manifest, uri)

    if path:
        manifest["meta_brownie"]["registry_address"] = address
        # save a local copy before returning
        for subfolder in list(path.parents)[2::-1]:
            subfolder.mkdir(exist_ok=True)
        with path.open("w") as fp:
            json.dump(manifest, fp)

    return manifest


def process_manifest(manifest: Dict, uri: Optional[str] = None) -> Dict:

    """
    Processes a manifest for use with Brownie.

    Args:
        manifest: ethPM manifest
        uri: IPFS uri of the package
    """

    if manifest["manifest_version"] != "2":
        raise InvalidManifest(
            f"Brownie only supports v2 ethPM manifests, this "
            f"manifest is v{manifest['manifest_version']}"
        )

    for key in ("contract_types", "deployments", "sources"):
        manifest.setdefault(key, {})

    # resolve sources
    for key in list(manifest["sources"]):
        content = manifest["sources"].pop(key)
        if _is_uri(content):
            content = _get_uri_contents(content)
        # ensure all absolute imports begin with contracts/
        content = re.sub(
            r"""(import((\s*{[^};]*}\s*from)|)\s*)("|')(contracts/||/)(?=[^./])""",
            lambda k: f"{k.group(1)}{k.group(4)}contracts/",
            content,
        )
        path = Path("/").joinpath(key.lstrip("./")).resolve()
        path_str = path.as_posix()[len(path.anchor) :]
        manifest["sources"][f"contracts/{path_str}"] = content

    # set contract_name in contract_types
    contract_types = manifest["contract_types"]
    for key, value in contract_types.items():
        if "contract_name" not in value:
            value["contract_name"] = key

    # resolve package dependencies
    for dependency_uri in manifest.pop("build_dependencies", {}).values():
        dep_manifest = get_manifest(dependency_uri)
        for key in ("sources", "contract_types"):
            for k in [i for i in manifest[key] if i in dep_manifest[key]]:
                if manifest[key][k] != dep_manifest[key][k]:
                    raise InvalidManifest("Namespace collision between package dependencies")
            manifest[key].update(dep_manifest[key])

    # compile sources to expand contract_types
    if manifest["sources"]:
        version = compiler.find_best_solc_version(manifest["sources"], install_needed=True)

        build_json = compiler.compile_and_format(manifest["sources"], version)
        for key, build in build_json.items():
            manifest["contract_types"].setdefault(key, {"contract_name": key})
            manifest["contract_types"][key].update(
                {
                    "abi": build["abi"],
                    "source_path": build["sourcePath"],
                    "all_source_paths": build["allSourcePaths"],
                    "compiler": build["compiler"],
                }
            )

    # delete contract_types with no source or ABI, we can't do much with them
    manifest["contract_types"] = dict(
        (k, v) for k, v in manifest["contract_types"].items() if "abi" in v
    )

    # resolve or delete deployments
    for chain_uri in list(manifest["deployments"]):
        deployments = manifest["deployments"][chain_uri]
        for name in list(deployments):
            deployments[name]["address"] = to_address(deployments[name]["address"])
            alias = deployments[name]["contract_type"]
            alias = alias[alias.rfind(":") + 1 :]
            deployments[name]["contract_type"] = alias
            if alias not in manifest["contract_types"]:
                del deployments[name]
        if not deployments:
            del manifest["deployments"][chain_uri]

    manifest["meta_brownie"] = {"manifest_uri": uri, "registry_address": None}
    return manifest


def _is_uri(uri: str) -> bool:
    try:
        result = urlparse(uri)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def _get_uri_contents(uri: str) -> str:
    path = _get_data_folder().joinpath(f"ipfs_cache/{urlparse(uri).netloc}.ipfs")
    path.parent.mkdir(exist_ok=True)
    if not path.exists():
        data = InfuraIPFSBackend().fetch_uri_contents(uri)
        with path.open("wb") as fp:
            fp.write(data)
        return data.decode("utf-8")
    with path.open() as fp:
        data = fp.read()
    return data


def get_deployment_addresses(
    manifest: Dict, contract_name: str, genesis_hash: Optional[str] = None
) -> List:

    """
    Parses a manifest and returns a list of deployment addresses for the given contract
    and chain.

    Args:
        manifest: ethPM manifest
        contract_name: Name of the contract
        genesis_block: Genesis block hash for the chain to return deployments on. If
                       None, the currently active chain will be used.
    """

    if genesis_hash is None:
        genesis_hash = web3.genesis_hash

    if "meta_brownie" not in manifest:
        manifest = process_manifest(manifest)

    chain_uri = f"blockchain://{genesis_hash}"
    key = next((i for i in manifest["deployments"] if i.startswith(chain_uri)), None)
    if key is None:
        return []
    return [
        v["address"]
        for v in manifest["deployments"][key].values()
        if manifest["contract_types"][v["contract_type"]]["contract_name"] == contract_name
    ]


def get_installed_packages(project_path: Path) -> Tuple[List, List]:

    """
    Returns a list of a installed ethPM packages within a project, and a list
    of packages that are installed and one or more files are modified or deleted.

    Args:
        project_path: Path to the root folder of the project

    Returns:
        (project name, version) of installed packages
        (project name, version) of installed-but-modified packages
    """

    packages_json = _load_packages_json(project_path)

    # determine if packages are installed, modified, or deleted
    installed: Set = set(packages_json["packages"])
    modified: Set = set()
    deleted: Set = set(packages_json["packages"])
    for source_path in list(packages_json["sources"]):
        package_list = packages_json["sources"][source_path]["packages"]

        # source does not exist, package has been modified
        if not project_path.joinpath(source_path).exists():
            installed.difference_update(package_list)
            modified.update(package_list)
            continue

        # source exists, package has NOT been deleted
        deleted.difference_update(package_list)
        with project_path.joinpath(source_path).open("rb") as fp:
            source = fp.read()

        if hashlib.md5(source).hexdigest() != packages_json["sources"][source_path]["md5"]:
            # package has been modified
            modified.update(package_list)

    # deleted packages have not been modified, modified packages have not been deleted
    modified.difference_update(deleted)
    installed.difference_update(modified)

    # properly remove deleted packages
    for package_name in deleted:
        remove_package(project_path, package_name, True)

    return (
        [(i, packages_json["packages"][i]["version"]) for i in sorted(installed)],
        [(i, packages_json["packages"][i]["version"]) for i in sorted(modified)],
    )


def install_package(project_path: Path, uri: str, replace_existing: bool = False) -> str:

    """
    Installs an ethPM package within the project.

    Args:
        project_path: Path to the root folder of the project
        uri: manifest URI, can be erc1319 or ipfs
        replace_existing: if True, existing files will be overwritten when
                            installing the package

    Returns: Name of the package
    """

    manifest = get_manifest(uri)
    package_name = manifest["package_name"]
    remove_package(project_path, package_name, True)
    packages_json = _load_packages_json(project_path)

    for path, source in manifest["sources"].items():
        source_path = project_path.joinpath(path)
        if not replace_existing and source_path.exists():
            with source_path.open() as fp:
                if fp.read() != source:
                    raise FileExistsError(
                        f"Cannot overwrite existing file with different content: '{source_path}'"
                    )

    for path, source in manifest["sources"].items():
        for folder in list(Path(path).parents)[::-1]:
            project_path.joinpath(folder).mkdir(exist_ok=True)
        with project_path.joinpath(path).open("w") as fp:
            fp.write(source)
        with project_path.joinpath(path).open("rb") as fp:
            source_bytes = fp.read()

        packages_json["sources"].setdefault(path, {"packages": []})
        packages_json["sources"][path]["md5"] = hashlib.md5(source_bytes).hexdigest()
        packages_json["sources"][path]["packages"].append(package_name)

    packages_json["packages"][package_name] = {
        "manifest_uri": manifest["meta_brownie"]["manifest_uri"],
        "registry_address": manifest["meta_brownie"]["registry_address"],
        "version": manifest["version"],
    }

    with project_path.joinpath("build/packages.json").open("w") as fp:
        json.dump(packages_json, fp, indent=2, sort_keys=True)
    return manifest["package_name"]


def remove_package(project_path: Path, package_name: str, delete_files: bool) -> bool:

    """
    Removes an ethPM package from a project.

    Args:
        project_path: Path to the root folder of the project
        package_name: name of the package
        delete_files: if True, source files related to the package are deleted.
                      files that are still required by other installed packages
                      will not be deleted.

    Returns: boolean indicating if package was installed.
    """

    packages_json = _load_packages_json(project_path)
    if package_name not in packages_json["packages"]:
        return False
    del packages_json["packages"][package_name]

    for source_path in [
        k for k, v in packages_json["sources"].items() if package_name in v["packages"]
    ]:
        packages_json["sources"][source_path]["packages"].remove(package_name)
        if delete_files and not packages_json["sources"][source_path]["packages"]:
            # if source file is not associated with any other projects, delete it
            del packages_json["sources"][source_path]
            if project_path.joinpath(source_path).exists():
                project_path.joinpath(source_path).unlink()

            # remove empty folders
            for path in list(Path(source_path).parents)[:-2]:
                parent_path = project_path.joinpath(path)
                if parent_path.exists() and not list(parent_path.glob("*")):
                    parent_path.rmdir()

    with project_path.joinpath("build/packages.json").open("w") as fp:
        json.dump(packages_json, fp, indent=2, sort_keys=True)
    return True


def create_manifest(
    project_path: Path, package_config: Dict, pin_assets: bool = False, silent: bool = True
) -> Tuple[Dict, str]:

    """
    Creates a manifest from a project, and optionally pins it to IPFS.

    Arguments:
        project_path: Path to the root folder of the project
        package_config: Configuration settings for the manifest
        pin_assets: if True, all source files and the manifest will
                    be uploaded onto IPFS via Infura.

    Returns: generated manifest, ipfs uri of manifest
    """

    package_config = _remove_empty_fields(package_config)
    _verify_package_name(package_config["package_name"])

    if pin_assets:
        ipfs_backend = InfuraIPFSBackend()

    manifest = {
        "manifest_version": "2",
        "package_name": package_config["package_name"],
        "version": package_config["version"],
        "sources": {},
        "contract_types": {},
    }
    if "meta" in package_config:
        manifest["meta"] = package_config["meta"]

    # load packages.json and add build_dependencies
    packages_json: Dict = {"sources": {}, "packages": {}}
    if not package_config["settings"]["include_dependencies"]:
        installed, modified = get_installed_packages(project_path)
        if modified:
            raise InvalidManifest(
                f"Dependencies have been modified locally: {', '.join([i[0] for i in modified])}"
            )
        if installed:
            packages_json = _load_packages_json(project_path)
            manifest["build_dependencies"] = dict(
                (k, v["manifest_uri"]) for k, v in packages_json["packages"].items()
            )

    # add sources
    contract_path = project_path.joinpath("contracts")
    for path in contract_path.glob("**/*.sol"):
        if path.relative_to(project_path).as_posix() in packages_json["sources"]:
            continue
        if pin_assets:
            if not silent:
                print(f'Pinning "{color("bright magenta")}{path.name}{color}"...')
            uri = ipfs_backend.pin_assets(path)[0]["Hash"]
        else:
            with path.open("rb") as fp:
                uri = generate_file_hash(fp.read())
        manifest["sources"][f"./{path.relative_to(contract_path).as_posix()}"] = f"ipfs://{uri}"

    # add contract_types
    for path in project_path.glob("build/contracts/*.json"):
        with path.open() as fp:
            build_json = json.load(fp)
        if not build_json["bytecode"]:
            # skip contracts that cannot deploy
            continue
        if build_json["sourcePath"] in packages_json["sources"]:
            # skip dependencies
            continue
        manifest["contract_types"][build_json["contractName"]] = _get_contract_type(build_json)

    # add deployments
    deployment_networks = package_config["settings"]["deployment_networks"]
    if deployment_networks:
        active_network = network.show_active()
        if active_network:
            network.disconnect()
        manifest["deployments"] = {}
        if isinstance(deployment_networks, str):
            deployment_networks = [deployment_networks]
        if deployment_networks == ["*"]:
            deployment_networks = [i.stem for i in project_path.glob("build/deployments/*")]
        for network_name in deployment_networks:
            instances = list(project_path.glob(f"build/deployments/{network_name}/*.json"))
            if not instances:
                continue
            instances.sort(key=lambda k: k.stat().st_mtime, reverse=True)
            network.connect(network_name)
            manifest["deployments"][web3.chain_uri] = {}
            for path in instances:
                with path.open() as fp:
                    build_json = json.load(fp)

                alias = build_json["contractName"]
                source_path = build_json["sourcePath"]
                if source_path in packages_json["sources"]:
                    alias = f"{packages_json['sources'][source_path]['packages'][0]}:{alias}"

                if alias in manifest["contract_types"]:
                    # skip deployment if bytecode does not match that of contract_type
                    bytecode = manifest["contract_types"][alias]["deployment_bytecode"]["bytecode"]
                    if f"0x{build_json['bytecode']}" != bytecode:
                        continue
                else:
                    # add contract_type for dependency
                    manifest["contract_types"][alias] = _get_contract_type(build_json)

                key = build_json["contractName"]
                for i in itertools.count(1):
                    if key not in manifest["deployments"][web3.chain_uri]:
                        break
                    key = f"{build_json['contractName']}-{i}"

                manifest["deployments"][web3.chain_uri][key] = {
                    "address": path.stem,
                    "contract_type": alias,
                }
            network.disconnect()
        if active_network:
            network.connect(active_network)
        if not manifest["deployments"]:
            del manifest["deployments"]

    uri = None
    if pin_assets:
        if not silent:
            print("Pinning manifest...")
        temp_path = Path(tempfile.gettempdir()).joinpath("manifest.json")
        with temp_path.open("w") as fp:
            json.dump(manifest, fp, sort_keys=True, separators=(",", ":"))
        uri = ipfs_backend.pin_assets(temp_path)[0]["Hash"]

    return manifest, uri


def verify_manifest(package_name: str, version: str, uri: str) -> None:

    """
    Verifies the validity of a package at a given IPFS URI.

    Arguments:
        package_name: Package name
        version: Package version
        uri: IPFS uri

    Returns None if the package is valid, raises InvalidManifest if not.
    """

    _verify_package_name(package_name)
    data = InfuraIPFSBackend().fetch_uri_contents(uri).decode("utf-8")
    try:
        manifest = json.loads(data)
    except Exception:
        raise InvalidManifest("URI did not return valid JSON encoded data")
    if json.dumps(manifest, sort_keys=True, separators=(",", ":")) != data:
        raise InvalidManifest("JSON data is not tightly packed with sorted keys")
    for key, value in [
        ("manifest_version", "2"),
        ("package_name", package_name),
        ("version", version),
    ]:
        if manifest.get(key, None) != value:
            raise InvalidManifest(f"Missing or invalid field: {key}")
    try:
        process_manifest(manifest)
    except Exception as e:
        raise InvalidManifest(f"Cannot process manifest - {str(e)}")


def release_package(
    registry_address: str, account: AccountsType, package_name: str, version: str, uri: str
) -> TransactionReceiptType:

    """
    Creates a new release of a package at an ERC1319 registry.

    Arguments:
        registry_address: Address of the registry
        account: Account object used to broadcast the transaction to the registry
        package_name: Name of the package
        version: Package version
        uri: IPFS uri of the package
    """

    registry = network.contract.Contract(
        "ERC1319Registry", registry_address, REGISTRY_ABI, owner=account
    )
    verify_manifest(package_name, version, uri)
    return registry.release(package_name, version, uri)


def _get_contract_type(build_json: Dict) -> Dict:
    contract_type = {
        "contract_name": build_json["contractName"],
        "source_path": f"./{Path(build_json['sourcePath']).relative_to('contracts')}",
        "deployment_bytecode": {"bytecode": f"0x{build_json['bytecode']}"},
        "runtime_bytecode": {"bytecode": f"0x{build_json['deployedBytecode']}"},
        "abi": build_json["abi"],
        "compiler": {
            "name": "solc" if build_json["language"] == "Solidity" else "vyper",
            "version": build_json["compiler"]["version"],
            "settings": {"evmVersion": build_json["compiler"]["evm_version"]},
        },
    }
    if build_json["language"] == "Solidity":
        contract_type["compiler"]["settings"]["optimizer"] = {
            "enabled": build_json["compiler"]["optimize"],
            "runs": build_json["compiler"]["runs"],
        }
    return contract_type


def _load_packages_json(project_path: Path) -> Dict:
    try:
        with project_path.joinpath("build/packages.json").open() as fp:
            return json.load(fp)
    except (FileNotFoundError, json.decoder.JSONDecodeError):
        return {"sources": {}, "packages": {}}


def _remove_empty_fields(initial: Dict) -> Dict:
    result = {}
    for key, value in initial.items():
        if isinstance(initial[key], dict):
            value = _remove_empty_fields(value)
        if isinstance(initial[key], list):
            value = [i for i in initial[key] if i is not None]
        if value not in (None, {}, [], ""):
            result[key] = value
    return result


def _verify_package_name(package_name: str) -> None:
    if re.fullmatch(r"^[a-z][a-z0-9_-]{0,255}$", package_name) is None:
        raise ValueError(f"Invalid package name '{package_name}'")



#!/usr/bin/python3

from typing import List


def expand_source_map(source_map_str: str) -> List:
    # Expands the compressed sourceMap supplied by solc into a list of lists
    source_map: List = [_expand_row(i) if i else None for i in source_map_str.split(";")]
    for i, value in enumerate(source_map[1:], 1):
        if value is None:
            source_map[i] = source_map[i - 1]
            continue
        for x in range(4):
            if source_map[i][x] is None:
                source_map[i][x] = source_map[i - 1][x]
    return source_map


def _expand_row(row: str) -> List:
    result: List = [None] * 4
    # ignore the new "modifier depth" value in solidity 0.6.0
    for i, value in enumerate(row.split(":")[:4]):
        if value:
            result[i] = value if i == 3 else int(value)
    return result



#!/usr/bin/python3

import atexit
import code
import importlib
import sys

from docopt import docopt

import brownie
from brownie import network, project
from brownie._config import ARGV, BROWNIE_FOLDER, CONFIG, _update_argv_from_docopt
from brownie.utils import color

if sys.platform == "win32":
    from pyreadline import Readline

    readline = Readline()
else:
    import readline  # noqa: F401

__doc__ = f"""Usage: brownie console [options]

Options:
  --network <name>        Use a specific network (default {CONFIG['network']['default']})
  --tb -t                 Show entire python traceback on exceptions
  --help -h               Display this message

Connects to the network and opens the brownie console.
"""


def main():
    args = docopt(__doc__)
    _update_argv_from_docopt(args)

    if project.check_for_project():
        active_project = project.load()
        active_project.load_config()
        print(f"{active_project._name} is the active project.")
    else:
        active_project = None
        print("No project was loaded.")

    network.connect(ARGV["network"])

    shell = Console(active_project)
    shell.interact(banner="Brownie environment is ready.", exitmsg="")


class Console(code.InteractiveConsole):
    def __init__(self, project=None):
        locals_dict = dict((i, getattr(brownie, i)) for i in brownie.__all__)
        locals_dict["dir"] = self._dir

        # only make GUI available if Tkinter is installed
        try:
            Gui = importlib.import_module("brownie._gui").Gui
            locals_dict["Gui"] = Gui
        except ModuleNotFoundError:
            pass

        if project:
            project._update_and_register(locals_dict)
            history_file = project._path
        else:
            history_file = BROWNIE_FOLDER

        history_file = str(history_file.joinpath(".history").absolute())
        atexit.register(_atexit_readline, history_file)
        try:
            readline.read_history_file(history_file)
        except (FileNotFoundError, OSError):
            pass
        super().__init__(locals_dict)

    # console dir method, for simplified and colorful output
    def _dir(self, obj=None):
        if obj is None:
            results = [(k, v) for k, v in self.locals.items() if not k.startswith("_")]
        elif hasattr(obj, "__console_dir__"):
            results = [(i, getattr(obj, i)) for i in obj.__console_dir__]
        else:
            results = [(i, getattr(obj, i)) for i in dir(obj) if not i.startswith("_")]
        results = sorted(results, key=lambda k: k[0])
        self.write(f"[{f'{color}, '.join(_dir_color(i[1]) + i[0] for i in results)}{color}]\n")

    def _console_write(self, obj):
        text = repr(obj)
        try:
            if obj and isinstance(obj, dict):
                text = color.pretty_dict(obj)
            elif obj and isinstance(obj, (tuple, list, set)):
                text = color.pretty_sequence(obj)
        except (SyntaxError, NameError):
            pass
        print(text)

    def showsyntaxerror(self, filename):
        tb = color.format_tb(sys.exc_info()[1])
        self.write(tb + "\n")

    def showtraceback(self):
        tb = color.format_tb(sys.exc_info()[1], start=1)
        self.write(tb + "\n")

    # save user input to readline history file, filter for private keys
    def push(self, line):
        try:
            cls_, method = line[: line.index("(")].split(".")
            method = getattr(self.locals[cls_], method)
            if hasattr(method, "_private"):
                readline.replace_history_item(
                    readline.get_current_history_length() - 1, line[: line.index("(")] + "()"
                )
        except (ValueError, AttributeError, KeyError):
            pass
        return super().push(line)

    def runsource(self, source, filename="<input>", symbol="single"):
        try:
            code = self.compile(source, filename, "single")
        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(filename)
            return False

        if code is None:
            return True

        try:
            self.compile(source, filename, "eval")
            code = self.compile(f"__ret_value__ = {source}", filename, "exec")
        except Exception:
            pass
        self.runcode(code)
        if "__ret_value__" in self.locals and self.locals["__ret_value__"] is not None:
            self._console_write(self.locals["__ret_value__"])
            del self.locals["__ret_value__"]
        return False


def _dir_color(obj):
    if type(obj).__name__ == "module":
        return color("module")
    if hasattr(obj, "_dir_color"):
        return color(obj._dir_color)
    if not callable(obj):
        return color("value")
    return color("callable")


def _atexit_readline(history_file):
    readline.set_history_length(1000)
    readline.write_history_file(history_file)



#!/usr/bin/python3

import json
import re
import shutil
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml

from brownie._singleton import _Singleton

BROWNIE_FOLDER = Path(__file__).parent
DATA_FOLDER = Path.home().joinpath(".brownie")

REPLACE = ["active_network", "networks"]


class ConfigDict(dict):
    """Dict subclass that prevents adding new keys when locked"""

    def __init__(self, values: Dict = {}) -> None:
        self._locked = False
        super().__init__()
        self.update(values)

    def __setitem__(self, key: str, value: Any) -> None:
        if self._locked and key not in self:
            raise KeyError(f"{key} is not a known config setting")
        if type(value) is dict:
            value = ConfigDict(value)
        super().__setitem__(key, value)

    def update(self, arg):  # type: ignore
        for k, v in arg.items():
            self.__setitem__(k, v)

    def _lock(self) -> None:
        """Locks the dict so that new keys cannot be added"""
        for v in [i for i in self.values() if type(i) is ConfigDict]:
            v._lock()
        self._locked = True

    def _unlock(self) -> None:
        """Unlocks the dict so that new keys can be added"""
        for v in [i for i in self.values() if type(i) is ConfigDict]:
            v._unlock()
        self._locked = False

    def _copy(self) -> Dict:
        config_copy = {}
        for key, value in self.items():
            if isinstance(value, ConfigDict):
                value = value._copy()
            config_copy[key] = value
        return config_copy


def _get_project_config_path(project_path: Path):
    if project_path.is_dir():
        path = project_path.joinpath("brownie-config")
    else:
        path = project_path
    suffix = next((i for i in (".yml", ".yaml", ".json") if path.with_suffix(i).exists()), None)
    if suffix is not None:
        return path.with_suffix(suffix)
    return None


def _load_config(project_path: Path) -> Dict:
    # Loads configuration data from a file, returns as a dict
    path = _get_project_config_path(project_path)
    if path is None:
        raise ValueError("Project does not exist!")
    with path.open() as fp:
        if path.suffix in (".yaml", ".yml"):
            return yaml.safe_load(fp)
        raw_json = fp.read()
    valid_json = re.sub(r'\/\/[^"]*?(?=\n|$)', "", raw_json)
    return json.loads(valid_json)


def _load_default_config() -> "ConfigDict":
    # Loads the default configuration settings from brownie/data/config.yaml
    base_config = BROWNIE_FOLDER.joinpath("data/brownie-config.yaml")
    default_config = DATA_FOLDER.joinpath("brownie-config.yaml")

    if not DATA_FOLDER.exists():
        DATA_FOLDER.mkdir()
    if not default_config.exists():
        shutil.copy(base_config, default_config)

    config = _Singleton("Config", (ConfigDict,), {})(_load_config(base_config))  # type: ignore
    _recursive_update(config, _load_config(default_config), [])
    config["active_network"] = {"name": None}
    return config


def _load_project_config(project_path: Path) -> None:
    # Loads configuration settings from a project's brownie-config.yaml
    config_data = _load_config(project_path.joinpath("brownie-config"))
    CONFIG._unlock()
    _recursive_update(CONFIG, config_data, [])
    CONFIG.setdefault("active_network", {"name": None})
    CONFIG._lock()


def _load_project_compiler_config(project_path: Optional[Path]) -> Dict:
    if not project_path:
        return CONFIG["compiler"]
    compiler_data = _load_config(project_path.joinpath("brownie-config"))["compiler"]
    for key in [i for i in ("evm_version", "minify_source") if i not in compiler_data]:
        compiler_data[key] = compiler_data["solc"].pop(key)
    return compiler_data


def _modify_network_config(network: str = None) -> Dict:
    """Modifies the 'active_network' configuration settings"""
    CONFIG._unlock()
    try:
        if not network:
            network = CONFIG["network"]["default"]

        CONFIG["active_network"] = {
            **CONFIG["network"]["settings"],
            **CONFIG["network"]["networks"][network],
        }
        CONFIG["active_network"]["name"] = network

        if ARGV["cli"] == "test":
            CONFIG["active_network"].update(CONFIG["pytest"])
            if not CONFIG["active_network"]["reverting_tx_gas_limit"]:
                print("WARNING: Reverting transactions will NOT be broadcasted.")
        return CONFIG["active_network"]
    except KeyError:
        raise KeyError(f"Network '{network}' is not defined in config.json")
    finally:
        CONFIG._lock()


def _recursive_update(original: Dict, new: Dict, base: List) -> None:
    # merges project config with brownie default config
    for k in new:
        if type(new[k]) is dict and k in REPLACE:
            original[k] = new[k]
        elif type(new[k]) is dict and k in original:
            _recursive_update(original[k], new[k], base + [k])
        else:
            original[k] = new[k]


def _update_argv_from_docopt(args: Dict) -> None:
    ARGV.update(dict((k.lstrip("-"), v) for k, v in args.items()))


def _get_data_folder() -> Path:
    return DATA_FOLDER


# create argv object
ARGV = _Singleton("Argv", (defaultdict,), {})(lambda: None)  # type: ignore

# load config
CONFIG = _load_default_config()



from typing import TypeVar

from brownie.network.account import Accounts
from brownie.network.transaction import TransactionReceipt

# NETWORK
# Account
AccountsType = TypeVar("AccountsType", bound=Accounts)

# Transactions
TransactionReceiptType = TypeVar("TransactionReceiptType", bound=TransactionReceipt)



#!/usr/bin/python3

import tkinter as tk
from tkinter import ttk

from .styles import BUTTON_STYLE


class ToggleButton(tk.Button):
    def __init__(self, parent, text, keybind=None):
        self.active = False
        super().__init__(parent, text=text, command=self.toggle)
        self.root = self._root()
        self.configure(**BUTTON_STYLE)
        if keybind:
            self.root.bind(keybind, self.toggle)

    def toggle(self, event=None):
        if self.active:
            self.toggle_off()
            self.configure(relief="raised", background="#272727")
        else:
            if not self.toggle_on():
                return
            self.configure(relief="sunken", background="#383838")
        self.active = not self.active

    def toggle_on(self):
        pass

    def toggle_off(self):
        pass


class SelectBox(ttk.Combobox):
    def __init__(self, parent, initial, values):
        super().__init__(parent, state="readonly", font=(None, 16))
        self.root = self._root()
        self["values"] = sorted(values)
        self.set(initial)
        self.bind("<<ComboboxSelected>>", self._select)
        self.configure(width=1)

    def _select(self):
        value = self.get()
        self.selection_clear()
        return value



#!/usr/bin/python3

from copy import deepcopy
from typing import Dict

_coverage_eval: Dict[str, Dict] = {}
_cached: Dict[str, Dict] = {}
_active_txhash: set = set()


def get_coverage_eval():
    """Returns all coverage data, active and cached."""
    return {**_cached, **_coverage_eval}


def get_merged_coverage_eval():
    """Merges and returns all active coverage data as a single dict.

    Returns: coverage eval dict.
    """
    if not _coverage_eval:
        return {}
    coverage_eval_list = list(_coverage_eval.values())
    merged_eval = deepcopy(coverage_eval_list.pop())
    for coverage_eval in coverage_eval_list:
        for name in coverage_eval:
            if name not in merged_eval:
                merged_eval[name] = coverage_eval[name]
                continue
            for path, map_ in coverage_eval[name].items():
                if path not in merged_eval[name]:
                    merged_eval[name][path] = map_
                    continue
                for i in range(3):
                    merged_eval[name][path][i] = set(merged_eval[name][path][i]).union(map_[i])
    return merged_eval


def clear():
    """Clears all coverage eval data."""
    _coverage_eval.clear()
    _cached.clear()
    _active_txhash.clear()


def _add_transaction(txhash, coverage_eval):
    # Adds coverage eval data
    _coverage_eval[txhash] = coverage_eval
    _active_txhash.add(txhash)


def _add_cached_transaction(txhash, coverage_eval):
    # Adds coverage data to the cache
    _cached[txhash] = coverage_eval


def _check_cached(txhash, active=True):
    # Checks if a tx hash is present within the cache, and if yes adds it to the active data
    if txhash in _cached:
        _coverage_eval[txhash] = _cached.pop(txhash)
        if active:
            _active_txhash.add(txhash)
    return txhash in _coverage_eval


def _get_active_txlist():
    # Returns a list of coverage hashes that are currently marked as active
    return sorted(_active_txhash)


def _clear_active_txlist():
    # Clears the active coverage hash list
    _active_txhash.clear()



#!/usr/bin/python3

from typing import Dict, ItemsView, List, Optional, Sequence, Tuple, Union

from .sources import Sources, highlight_source

BUILD_KEYS = [
    "abi",
    "allSourcePaths",
    "ast",
    "bytecode",
    "bytecodeSha1",
    "compiler",
    "contractName",
    "coverageMap",
    "deployedBytecode",
    "deployedSourceMap",
    "dependencies",
    "language",
    "offset",
    "opcodes",
    "pcMap",
    "sha1",
    "source",
    "sourceMap",
    "sourcePath",
    "type",
]

_revert_map: Dict = {}


class Build:

    """Methods for accessing and manipulating a project's contract build data."""

    def __init__(self, sources: Sources) -> None:
        self._sources = sources
        self._build: Dict = {}

    def _add(self, build_json: Dict) -> None:
        contract_name = build_json["contractName"]
        if "0" in build_json["pcMap"]:
            build_json["pcMap"] = dict((int(k), v) for k, v in build_json["pcMap"].items())
        if build_json["compiler"]["minify_source"]:
            build_json = self.expand_build_offsets(build_json)
        self._build[contract_name] = build_json
        self._generate_revert_map(build_json["pcMap"], build_json["language"])

    def _generate_revert_map(self, pcMap: Dict, language: str) -> None:
        # Adds a contract's dev revert strings to the revert map and it's pcMap
        marker = "//" if language == "Solidity" else "#"
        for pc, data in (
            (k, v)
            for k, v in pcMap.items()
            if v["op"] in ("REVERT", "INVALID") or "jump_revert" in v
        ):
            if "dev" not in data:
                if "fn" not in data or "first_revert" in data:
                    _revert_map[pc] = False
                    continue
                try:
                    revert_str = self._sources.get(data["path"])[data["offset"][1] :]
                    revert_str = revert_str[: revert_str.index("\n")]
                    revert_str = revert_str[revert_str.index(marker) + len(marker) :].strip()
                    if revert_str.startswith("dev:"):
                        data["dev"] = revert_str
                except (KeyError, ValueError):
                    pass

            msg = "" if data["op"] == "REVERT" else "invalid opcode"
            revert = (
                data["path"],
                tuple(data["offset"]),
                data["fn"],
                data.get("dev", msg),
                self._sources,
            )

            # do not compare the final tuple item in case the same project was loaded twice
            if pc not in _revert_map or (_revert_map[pc] and revert[:-1] == _revert_map[pc][:-1]):
                _revert_map[pc] = revert
                continue
            _revert_map[pc] = False

    def _remove(self, contract_name: str) -> None:
        del self._build[self._stem(contract_name)]

    def get(self, contract_name: str) -> Dict:
        """Returns build data for the given contract name."""
        return self._build[self._stem(contract_name)]

    def items(self, path: Optional[str] = None) -> Union[ItemsView, List]:
        """Provides an list of tuples as (key,value), similar to calling dict.items.
        If a path is given, only contracts derived from that source file are returned."""
        if path is None:
            return self._build.items()
        return [(k, v) for k, v in self._build.items() if v["sourcePath"] == path]

    def contains(self, contract_name: str) -> bool:
        """Checks if the contract name exists in the currently loaded build data."""
        return self._stem(contract_name) in self._build

    def get_dependents(self, contract_name: str) -> List:
        """Returns a list of contract names that inherit from or link to the given
        contract. Used by the compiler when determining which contracts to recompile
        based on a changed source file."""
        return [k for k, v in self._build.items() if contract_name in v["dependencies"]]

    def _stem(self, contract_name: str) -> str:
        return contract_name.replace(".json", "")

    def expand_build_offsets(self, build_json: Dict) -> Dict:
        """Expands minified source offsets in a build json dict."""

        offset_map: Dict = {}
        name = build_json["contractName"]

        # minification only happens to the target contract that was compiled,
        # so we ignore any import sources
        source_path = build_json["sourcePath"]

        for value in (
            v
            for v in build_json["pcMap"].values()
            if "offset" in v and "path" in v and v["path"] == source_path
        ):
            value["offset"] = self._get_offset(offset_map, name, value["offset"])

        for key in ("branches", "statements"):
            coverage_map = build_json["coverageMap"][key][source_path]
            for fn, value in coverage_map.items():
                coverage_map[fn] = dict(
                    (k, self._get_offset(offset_map, name, v[:2]) + tuple(v[2:]))
                    for k, v in value.items()
                )
        return build_json

    def _get_offset(self, offset_map: Dict, name: str, offset: Sequence[int]) -> Tuple:
        offset = tuple(offset)
        if offset not in offset_map:
            offset_map[offset] = self._sources.expand_offset(name, offset)
        return offset_map[offset]


def _get_dev_revert(pc: int) -> Optional[str]:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the commented dev string (if any)
    if pc not in _revert_map or _revert_map[pc] is False:
        return None
    return _revert_map[pc][3]


def _get_error_source_from_pc(pc: int, pad: int = 3) -> Tuple:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the highlighted relevent source code and the method name.
    if pc not in _revert_map or _revert_map[pc] is False:
        return (None,) * 4
    revert = _revert_map[pc]
    source = revert[4].get(revert[0])
    highlight, linenos = highlight_source(source, revert[1], pad=pad)  # type: ignore
    return highlight, linenos, revert[0], revert[2]



#!/usr/bin/python3

import pytest

import brownie
from brownie._config import ARGV, CONFIG
from brownie.test import output
from brownie.test._manager import TestManager


class RevertContextManager:
    def __init__(self, revert_msg=None):
        self.revert_msg = revert_msg

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            raise AssertionError("Transaction did not revert") from None
        if exc_type is not brownie.exceptions.VirtualMachineError:
            raise exc_type(exc_value).with_traceback(traceback)
        if self.revert_msg is None or self.revert_msg == exc_value.revert_msg:
            return True
        raise AssertionError(
            f"Unexpected revert string '{exc_value.revert_msg}'\n{exc_value.source}"
        ) from None


def _generate_fixture(container):
    def _fixture():
        yield container

    _fixture.__doc__ = f"Provides access to Brownie ContractContainer object '{container._name}'"
    return pytest.fixture(scope="session")(_fixture)


if brownie.project.check_for_project("."):

    # load project and generate dynamic fixtures
    project = brownie.project.load()
    project.load_config()
    for container in project:
        globals()[container._name] = _generate_fixture(container)

    # create test manager - for reading and writing to build/test.json
    manager = TestManager(project)
    pytest.reverts = RevertContextManager

    # set commandline options
    def pytest_addoption(parser):
        parser.addoption(
            "--coverage", "-C", action="store_true", help="Evaluate contract test coverage"
        )
        parser.addoption(
            "--gas", "-G", action="store_true", help="Display gas profile for function calls"
        )
        parser.addoption(
            "--update", "-U", action="store_true", help="Only run tests where changes have occurred"
        )
        parser.addoption(
            "--revert-tb", "-R", action="store_true", help="Show detailed traceback on tx reverts"
        )
        parser.addoption(
            "--network",
            "-N",
            default=False,
            nargs=1,
            help=f"Use a specific network (default {CONFIG['network']['default']})",
        )

    def pytest_configure(config):
        for key in ("coverage", "always_transact"):
            ARGV[key] = config.getoption("--coverage")
        ARGV["gas"] = config.getoption("--gas")
        ARGV["revert"] = config.getoption("--revert-tb") or CONFIG["pytest"]["revert_traceback"]
        ARGV["update"] = config.getoption("--update")
        ARGV["network"] = None
        if config.getoption("--network"):
            ARGV["network"] = config.getoption("--network")[0]

    # plugin hooks

    def pytest_generate_tests(metafunc):
        # module_isolation always runs first
        fixtures = metafunc.fixturenames
        if "module_isolation" in fixtures:
            fixtures.remove("module_isolation")
            fixtures.insert(0, "module_isolation")
        # fn_isolation always runs before other function scoped fixtures
        if "fn_isolation" in fixtures:
            fixtures.remove("fn_isolation")
            defs = metafunc._arg2fixturedefs
            idx = next(
                (
                    fixtures.index(i)
                    for i in fixtures
                    if i in defs and defs[i][0].scope == "function"
                ),
                len(fixtures),
            )
            fixtures.insert(idx, "fn_isolation")

    def pytest_collection_modifyitems(items):
        # determine which modules are properly isolated
        tests = {}
        for i in items:
            if "skip_coverage" in i.fixturenames and ARGV["coverage"]:
                i.add_marker("skip")
            path = i.parent.fspath
            if "module_isolation" not in i.fixturenames:
                tests[path] = None
                continue
            if path in tests and tests[path] is None:
                continue
            tests.setdefault(i.parent.fspath, []).append(i)
        isolated_tests = sorted(k for k, v in tests.items() if v)
        manager.set_isolated_modules(isolated_tests)

        if ARGV["update"]:
            isolated_tests = sorted(filter(manager.check_updated, tests))
            # if all tests will be skipped, do not launch the rpc client
            if sorted(tests) == isolated_tests:
                ARGV["norpc"] = True
            # if update flag is active, add skip marker to unchanged tests
            for path in isolated_tests:
                tests[path][0].parent.add_marker("skip")

    def pytest_runtestloop():
        if not ARGV["norpc"]:
            brownie.network.connect(ARGV["network"])

    def pytest_runtest_protocol(item):
        manager.set_active(item.parent.fspath)

    def pytest_report_teststatus(report):
        return manager.check_status(report)

    def pytest_runtest_teardown(item, nextitem):
        if list(item.parent.iter_markers("skip")):
            return
        # if this is the last test in a module, record the results
        if not nextitem or item.parent.fspath != nextitem.parent.fspath:
            manager.module_completed(item.parent.fspath)

    def pytest_sessionfinish():
        manager.save_json()
        if ARGV["coverage"]:
            coverage_eval = brownie.test.coverage.get_merged_coverage_eval()
            output._print_coverage_totals(project._build, coverage_eval)
            output._save_coverage_report(
                project._build, coverage_eval, project._path.joinpath("reports")
            )
        if ARGV["gas"]:
            output._print_gas_profile()
        project.close(False)

    def pytest_keyboard_interrupt():
        ARGV["interrupt"] = True

    # fixtures
    @pytest.fixture(scope="module")
    def module_isolation():
        brownie.rpc.reset()
        yield
        if not ARGV["interrupt"]:
            brownie.rpc.reset()

    @pytest.fixture
    def fn_isolation(module_isolation):
        brownie.rpc.snapshot()
        yield
        if not ARGV["interrupt"]:
            brownie.rpc.revert()

    @pytest.fixture(scope="session")
    def a():
        yield brownie.accounts

    @pytest.fixture(scope="session")
    def accounts():
        yield brownie.accounts

    @pytest.fixture(scope="session")
    def history():
        yield brownie.history

    @pytest.fixture(scope="session")
    def rpc():
        yield brownie.rpc

    @pytest.fixture(scope="session")
    def web3():
        yield brownie.web3

    @pytest.fixture
    def no_call_coverage():
        ARGV["always_transact"] = False
        yield
        ARGV["always_transact"] = ARGV["coverage"]

    @pytest.fixture
    def skip_coverage():
        # implemented in pytest_collection_modifyitems
        pass



#!/usr/bin/python3

import json
import tkinter as tk
from tkinter import ttk

from brownie.project import get_loaded_projects

from .console import Console, ConsoleButton
from .opcodes import OpcodeList, ScopingButton
from .report import HighlightSelect, ReportSelect
from .source import ContractSelect, SourceNoteBook
from .styles import set_style
from .tooltip import ToolTip


class Root(tk.Tk):

    _active = False

    def __init__(self):
        projects = get_loaded_projects()

        if self._active:
            raise SystemError("GUI is already active")
        if not projects:
            raise SystemError("No project loaded")
        if len(projects) > 1:
            raise SystemError("More than one active project")
        Root._active = True

        self.active_project = projects[0]
        self.pcMap = {}

        self.report_key = None
        self.highlight_key = None
        self.reports = {}
        for path in self.active_project._path.glob("reports/*.json"):
            try:
                with path.open() as fp:
                    self.reports[path.stem] = json.load(fp)
            except Exception:
                continue

        super().__init__(className=f" Brownie GUI - {self.active_project._name}")
        self.bind("<Escape>", lambda k: self.destroy())

        # geometry and styling
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, minsize=30)
        self.rowconfigure(1, weight=1)
        set_style(self)

        # main widgets
        self.main = MainFrame(self)
        self.main.grid(row=1, column=0, sticky="nsew")

        # toolbar widgets
        self.toolbar = ToolbarFrame(self, self.active_project)
        self.toolbar.grid(row=0, column=0, sticky="nsew")

    @property
    def active_report(self):
        return self.reports[self.report_key]["highlights"]

    def get_active_contract(self):
        return self.toolbar.combo.get()

    def set_active_contract(self, contract_name):
        build_json = self.active_project._build.get(contract_name)
        self.main.note.set_visible(build_json["allSourcePaths"])
        self.main.note.set_active(build_json["sourcePath"])
        self.main.oplist.set_opcodes(build_json["pcMap"])
        self.pcMap = dict((str(k), v) for k, v in build_json["pcMap"].items())

    def destroy(self):
        super().destroy()
        self.quit()
        Root._active = False


class MainFrame(ttk.Frame):
    def __init__(self, root):
        super().__init__(root)

        # geometry
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, minsize=280)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, minsize=24)

        self.oplist = OpcodeList(self, (("pc", 80), ("opcode", 200)))
        self.oplist.grid(row=0, column=1, rowspan=2, sticky="nsew")

        self.note = SourceNoteBook(self)
        self.note.grid(row=0, column=0, sticky="nsew")

        self.console = Console(self)
        self.console.grid(row=1, column=0, sticky="nsew")


class ToolbarFrame(ttk.Frame):
    def __init__(self, root, project):
        super().__init__(root)
        self.root = root

        # geometry
        self.columnconfigure([0, 1], minsize=80)
        self.columnconfigure(7, weight=1)
        self.columnconfigure([8, 9], minsize=200)
        self.columnconfigure(10, minsize=304)

        # toggle buttons
        self.console = ConsoleButton(self)
        self.console.grid(row=0, column=0, sticky="nsew")
        ToolTip(self.console, "Toggle expanded console")

        self.scope = ScopingButton(self)
        self.scope.grid(row=0, column=1, sticky="nsew")
        ToolTip(self.scope, "Filter opcodes to only show those\nrelated to the highlighted source")

        # report selection
        self.highlight_select = HighlightSelect(self)
        self.highlight_select.grid(row=0, column=8, sticky="nsew", padx=10)
        self.highlight_select.hide()
        ToolTip(self.highlight_select, "Select which report highlights to display")

        self.report = ReportSelect(self)
        self.report.grid(row=0, column=9, sticky="nsew", padx=10)
        ToolTip(self.report, "Select a report to overlay onto the source code")

        # contract selection
        self.combo = ContractSelect(self, [k for k, v in project._build.items() if v["bytecode"]])
        self.combo.grid(row=0, column=10, sticky="nsew")
        ToolTip(self.combo, "Select the contract source to view")



#!/usr/bin/python3


import time
from tkinter import ttk

from brownie.project.sources import is_inside_offset

from .bases import ToggleButton


class OpcodeList(ttk.Treeview):
    def __init__(self, parent, columns, **kwargs):
        self._last = set()
        self._seek_buffer = ""
        self._seek_last = 0
        self._highlighted = set()
        self._frame = ttk.Frame(parent)
        super().__init__(
            self._frame, columns=[i[0] for i in columns[1:]], selectmode="browse", **kwargs
        )
        self.pack(side="left", fill="y")
        self.heading("#0", text=columns[0][0])
        self.column("#0", width=columns[0][1])
        for tag, width in columns[1:]:
            self.heading(tag, text=tag)
            self.column(tag, width=width)

        scroll = ttk.Scrollbar(self._frame)
        scroll.pack(side="right", fill="y")

        self.configure(yscrollcommand=scroll.set)
        scroll.configure(command=self.yview)

        self.tag_configure("NoSource", background="#161616")
        self.bind("<<TreeviewSelect>>", self._select_bind)

        root = self.root = self._root()
        root.bind("j", self._highlight_jumps)
        root.bind("r", self._highlight_revert)
        self.bind("<3>", self._highlight_opcode)
        for i in range(10):
            root.bind(str(i), self._seek)

    def grid(self, *args, **kwargs):
        self._frame.grid(*args, **kwargs)

    def insert(self, values, tags=[]):
        super().insert("", "end", iid=values[0], text=values[0], values=values[1:], tags=tags)

    def delete_all(self):
        for item in self.get_children():
            self.delete(item)

    def clear_selection(self):
        self.selection_remove(self.selection())

    def selection_set(self, id_):
        self.see(id_)
        super().selection_set(id_)
        self.focus_set()
        self.focus(id_)

    def set_opcodes(self, pcMap):
        self.delete_all()
        for pc, op in [(i, pcMap[i]) for i in sorted(pcMap)]:
            if "path" not in op or (
                op["path"] == pcMap[0]["path"] and op["offset"] == pcMap[0]["offset"]
            ):
                tag = "NoSource"
            else:
                # used to find all the opcodes with the same source offset
                tag = f"{op['path']}:{op['offset'][0]}:{op['offset'][1]}"
            self.insert([str(pc), op["op"]], [tag, op["op"], str(pc)])

    def _select_bind(self, event=None):
        for tag in self._last:
            self.tag_configure(tag, background="", foreground="")
        self._last.clear()
        try:
            pc = self.selection()[0]
        except IndexError:
            return
        pcMap = self.root.pcMap[pc]
        note = self.root.main.note

        console = self.root.main.console
        console.write(f"{pc} {pcMap['op']}")
        if "value" in pcMap:
            console.append(f" {pcMap['value']}")
        if "offset" in pcMap:
            console.append(f"\nOffsets: {pcMap['offset'][0]}, {pcMap['offset'][1]}")

        if pcMap["op"] in ("JUMP", "JUMPI"):
            prev = self._get_prev(pc)
            if self.root.pcMap[prev]["op"] in ("PUSH1", "PUSH2"):
                tag = int(self.root.pcMap[prev]["value"], 16)
                self.tag_configure(tag, foreground="#00ff00")
                self._last.add(tag)
                console.append(f"\nTarget: {tag}")

        if pcMap["op"] == "JUMPDEST":
            targets = _get_targets(self.root.pcMap, pc)
            if targets:
                console.append(f"\nJumps: {', '.join(targets)}")
            else:
                console.append(f"\nJumps: None")
            for item in targets:
                self.tag_configure(item, foreground="#00ff00")
                self._last.add(item)

        tag = self.item(pc, "tags")[0]
        if tag == "NoSource":
            note.active_frame().clear_highlight()
            return
        self.tag_configure(tag, background="#2a4864")
        self._last.add(tag)
        if "path" not in pcMap:
            note.active_frame().clear_highlight()
            return
        note.set_active(pcMap["path"])
        note.active_frame().highlight(*pcMap["offset"])

    def _seek(self, event):
        if self._seek_last < time.time() - 1:
            self._seek_buffer = ""
        self._seek_last = time.time()
        self._seek_buffer += event.char
        pc = sorted([int(i) for i in self.root.pcMap])[::-1]
        id_ = next(str(i) for i in pc if i <= int(self._seek_buffer))
        self.selection_set(id_)

    def _highlight_opcode(self, event):
        pc = self.identify_row(event.y)
        op = self.root.pcMap[pc]["op"]
        if op in self._highlighted:
            self.tag_configure(op, foreground="")
            self._highlighted.remove(op)
            return
        self.tag_configure(op, foreground="#dd3333" if op in ("REVERT", "INVALID") else "#dddd33")
        self._highlighted.add(op)

    def _highlight_jumps(self, event):
        try:
            pc = self.selection()[0]
        except IndexError:
            return
        if self.root.pcMap[pc]["op"] not in ("JUMP", "JUMPI"):
            return
        prev = self._get_prev(pc)
        if self.root.pcMap[prev]["op"] in ("PUSH1", "PUSH2"):
            tag = int(self.root.pcMap[prev]["value"], 16)
            self.see(tag)
            self._select_bind()

    def _highlight_revert(self, event):
        if "REVERT" in self._highlighted:
            self.tag_configure("REVERT", foreground="")
            self._highlighted.discard("REVERT")
        else:
            self.tag_configure("REVERT", foreground="#dd3333")
            self._highlighted.add("REVERT")

    def _get_prev(self, pc):
        pc_list = sorted(self.root.pcMap, key=lambda k: int(k), reverse=True)
        return next(i for i in pc_list if int(i) < int(pc))


class ScopingButton(ToggleButton):
    def __init__(self, parent):
        super().__init__(parent, "Scope", "s")
        self.oplist = self.root.main.oplist

    def toggle_on(self):
        try:
            op = self.oplist.selection()[0]
        except IndexError:
            return False
        if self.oplist.item(op, "tags")[0] == "NoSource":
            return False
        pc = self.root.pcMap[op]
        for key, value in sorted(self.root.pcMap.items(), key=lambda k: int(k[0])):
            if (
                "path" not in value
                or value["path"] != pc["path"]
                or not is_inside_offset(value["offset"], pc["offset"])
            ):
                self.oplist.detach(key)
            else:
                self.oplist.move(key, "", key)
        self.oplist.see(op)
        self.root.main.note.apply_scope(*pc["offset"])
        return True

    def toggle_off(self):
        self.root.main.note.clear_scope()
        for i in sorted(self.root.pcMap, key=lambda k: int(k)):
            self.oplist.move(i, "", i)
        if self.oplist.selection():
            self.oplist.see(self.oplist.selection()[0])


def _get_targets(pc_map, pc):
    targets = []
    for key, value in pc_map.items():
        if int(value.get("value", "0"), 16) != int(pc):
            continue
        next_pc = str(int(key) + int(value["op"][4:]) + 1)
        if pc_map[next_pc]["op"] in ("JUMP", "JUMPI"):
            targets.append(next_pc)
    return targets



#!/usr/bin/python3

import json
import shutil
import sys
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Iterator, KeysView, List, Optional, Set, Union

import requests

from brownie._config import (
    BROWNIE_FOLDER,
    CONFIG,
    _get_data_folder,
    _get_project_config_path,
    _load_project_compiler_config,
    _load_project_config,
)
from brownie.exceptions import ProjectAlreadyLoaded, ProjectNotFound
from brownie.network import web3
from brownie.network.contract import ContractContainer, ProjectContract
from brownie.network.state import _remove_contract
from brownie.project import compiler
from brownie.project.build import BUILD_KEYS, Build
from brownie.project.ethpm import get_deployment_addresses, get_manifest
from brownie.project.sources import Sources, get_hash
from brownie.utils import color

FOLDERS = [
    "contracts",
    "scripts",
    "reports",
    "tests",
    "build",
    "build/contracts",
    "build/deployments",
]
MIXES_URL = "https://github.com/brownie-mix/{}-mix/archive/master.zip"

_loaded_projects = []


class _ProjectBase:
    def __init__(self, name: str, contract_sources: Dict, project_path: Optional[Path]) -> None:
        self._path = project_path
        self._name = name
        self._sources = Sources(contract_sources)
        self._build = Build(self._sources)

    def _compile(self, sources: Dict, compiler_config: Dict, silent: bool) -> None:
        allow_paths = None
        if self._path is not None:
            allow_paths = self._path.joinpath("contracts").as_posix()
        compiler_config.setdefault("solc", {})
        build_json = compiler.compile_and_format(
            sources,
            solc_version=compiler_config["solc"].get("version", None),
            optimize=compiler_config["solc"].get("optimize", None),
            runs=compiler_config["solc"].get("runs", None),
            evm_version=compiler_config["evm_version"],
            minify=compiler_config["minify_source"],
            silent=silent,
            allow_paths=allow_paths,
        )
        for data in build_json.values():
            if self._path is not None:
                path = self._path.joinpath(f"build/contracts/{data['contractName']}.json")
                with path.open("w") as fp:
                    json.dump(data, fp, sort_keys=True, indent=2, default=sorted)
            self._build._add(data)

    def _create_containers(self) -> None:
        # create container objects
        self._containers: Dict = {}
        for key, data in self._build.items():
            if data["bytecode"]:
                container = ContractContainer(self, data)
                self._containers[key] = container
                setattr(self, container._name, container)

    def __getitem__(self, key: str) -> ContractContainer:
        return self._containers[key]

    def __iter__(self) -> Iterator[ContractContainer]:
        return iter(self._containers[i] for i in sorted(self._containers))

    def __len__(self) -> int:
        return len(self._containers)

    def __contains__(self, item: ContractContainer) -> bool:
        return item in self._containers

    def dict(self) -> Dict:
        return dict(self._containers)

    def keys(self) -> KeysView[Any]:
        return self._containers.keys()


class Project(_ProjectBase):

    """
    Top level dict-like container that holds data and objects related to
    a brownie project.

    Attributes:
        _path: Path object, absolute path to the project
        _name: Name that the project is loaded as
        _sources: project Source object
        _build: project Build object
    """

    def __init__(self, name: str, project_path: Path) -> None:
        contract_sources: Dict = {}
        for path in project_path.glob("contracts/**/*"):
            if "/_" in path.as_posix() or path.suffix not in (".sol", ".vy"):
                continue
            with path.open() as fp:
                source = fp.read()
            path_str: str = path.relative_to(project_path).as_posix()
            contract_sources[path_str] = source
        super().__init__(name, contract_sources, project_path)
        self._path: Path = project_path

        contract_list = self._sources.get_contract_list()
        for path in list(project_path.glob("build/contracts/*.json")):
            try:
                with path.open() as fp:
                    build_json = json.load(fp)
            except json.JSONDecodeError:
                build_json = {}
            if not set(BUILD_KEYS).issubset(build_json) or path.stem not in contract_list:
                path.unlink()
                continue
            self._build._add(build_json)

        self._active = False
        self.load()

    def load(self) -> None:
        """Compiles the project contracts, creates ContractContainer objects and
        populates the namespace."""
        if self._active:
            raise ProjectAlreadyLoaded("Project is already active")

        self._compiler_config = _load_project_compiler_config(self._path)

        # compile updated sources, update build
        changed = self._get_changed_contracts()
        self._compile(changed, self._compiler_config, False)
        self._create_containers()
        self._load_deployments()

        # add project to namespaces, apply import blackmagic
        name = self._name
        self.__all__ = list(self._containers)
        sys.modules[f"brownie.project.{name}"] = self  # type: ignore
        sys.modules["brownie.project"].__dict__[name] = self
        sys.modules["brownie.project"].__all__.append(name)  # type: ignore
        sys.modules["brownie.project"].__console_dir__.append(name)  # type: ignore
        self._namespaces = [
            sys.modules["__main__"].__dict__,
            sys.modules["brownie.project"].__dict__,
        ]
        self._active = True
        _loaded_projects.append(self)

    def _get_changed_contracts(self) -> Dict:
        changed = [i for i in self._sources.get_contract_list() if self._compare_build_json(i)]
        final = set(changed)
        for contract_name in changed:
            final.update(self._build.get_dependents(contract_name))
        for name in [i for i in final if self._build.contains(i)]:
            self._build._remove(name)
        changed_set: Set = set(self._sources.get_source_path(i) for i in final)
        return dict((i, self._sources.get(i)) for i in changed_set)

    def _compare_build_json(self, contract_name: str) -> bool:
        config = self._compiler_config
        try:
            source = self._sources.get(contract_name)
            build_json = self._build.get(contract_name)
        except KeyError:
            return True
        if build_json["sha1"] != get_hash(
            source, contract_name, config["minify_source"], build_json["language"]
        ):
            return True
        if _compare_settings(config, build_json["compiler"]):
            return True
        if build_json["language"] == "Solidity" and _compare_settings(
            config["solc"], build_json["compiler"]
        ):
            return True
        return False

    def _load_deployments(self) -> None:
        if not CONFIG["active_network"].get("persist", None):
            return
        network = CONFIG["active_network"]["name"]
        path = self._path.joinpath(f"build/deployments/{network}")
        path.mkdir(exist_ok=True)
        deployments = list(
            self._path.glob(f"build/deployments/{CONFIG['active_network']['name']}/*.json")
        )
        deployments.sort(key=lambda k: k.stat().st_mtime)
        for build_json in deployments:
            with build_json.open() as fp:
                build = json.load(fp)
            if build["contractName"] not in self._containers:
                build_json.unlink()
            else:
                container = self._containers[build["contractName"]]
                contract = ProjectContract(self, build, build_json.stem)
                container._contracts.append(contract)

    def _update_and_register(self, dict_: Any) -> None:
        dict_.update(self)
        self._namespaces.append(dict_)

    def __repr__(self) -> str:
        return f"<Project object '{color['string']}{self._name}{color}'>"

    def load_config(self) -> None:
        """Loads the project config file settings"""
        if isinstance(self._path, Path):
            _load_project_config(self._path)

    def close(self, raises: bool = True) -> None:
        """Removes pointers to the project's ContractContainer objects and this object."""
        if not self._active:
            if not raises:
                return
            raise ProjectNotFound("Project is not currently loaded.")

        # remove objects from namespace
        for dict_ in self._namespaces:
            for key in [
                k
                for k, v in dict_.items()
                if v == self or (k in self and v == self[k])  # type: ignore
            ]:
                del dict_[key]

        # remove contracts
        for contract in [x for v in self._containers.values() for x in v._contracts]:
            _remove_contract(contract)
        for container in self._containers.values():
            container._contracts.clear()
        self._containers.clear()

        # undo black-magic
        name = self._name
        del sys.modules[f"brownie.project.{name}"]
        sys.modules["brownie.project"].__all__.remove(name)  # type: ignore
        sys.modules["brownie.project"].__console_dir__.remove(name)  # type: ignore
        self._active = False
        _loaded_projects.remove(self)

        # clear paths
        try:
            sys.path.remove(str(self._path))
        except ValueError:
            pass


class TempProject(_ProjectBase):

    """Simplified Project class used to hold temporary contracts that are
    compiled via project.compile_source"""

    def __init__(self, name: str, contract_sources: Dict, compiler_config: Dict) -> None:
        super().__init__(name, contract_sources, None)
        self._compile(contract_sources, compiler_config, True)
        self._create_containers()

    def __repr__(self) -> str:
        return f"<TempProject object '{color['string']}{self._name}{color}'>"


def check_for_project(path: Union[Path, str] = ".") -> Optional[Path]:
    """Checks for a Brownie project."""
    path = Path(path).resolve()
    for folder in [path] + list(path.parents):
        if _get_project_config_path(folder):
            return folder
    return None


def get_loaded_projects() -> List:
    """Returns a list of currently loaded Project objects."""
    return _loaded_projects.copy()


def new(project_path_str: str = ".", ignore_subfolder: bool = False) -> str:
    """Initializes a new project.

    Args:
        project_path: Path to initialize the project at. If not exists, it will be created.
        ignore_subfolders: If True, will not raise if initializing in a project subfolder.

    Returns the path to the project as a string.
    """
    project_path = _new_checks(project_path_str, ignore_subfolder)
    project_path.mkdir(exist_ok=True)
    _create_folders(project_path)
    if not _get_project_config_path(project_path):
        shutil.copy(
            _get_data_folder().joinpath("brownie-config.yaml"),
            project_path.joinpath("brownie-config.yaml"),
        )
    if not project_path.joinpath("ethpm-config.yaml").exists():
        shutil.copy(
            BROWNIE_FOLDER.joinpath("data/ethpm-config.yaml"),
            project_path.joinpath("ethpm-config.yaml"),
        )
    _add_to_sys_path(project_path)
    return str(project_path)


def from_brownie_mix(
    project_name: str, project_path: Union[Path, str] = None, ignore_subfolder: bool = False
) -> str:
    """Initializes a new project via a template. Templates are downloaded from
    https://www.github.com/brownie-mix

    Args:
        project_path: Path to initialize the project at.
        ignore_subfolders: If True, will not raise if initializing in a project subfolder.

    Returns the path to the project as a string.
    """
    project_name = str(project_name).replace("-mix", "")
    url = MIXES_URL.format(project_name)
    if project_path is None:
        project_path = Path(".").joinpath(project_name)
    project_path = _new_checks(project_path, ignore_subfolder)
    if project_path.exists() and list(project_path.glob("*")):
        raise FileExistsError(f"Folder already exists - {project_path}")

    print(f"Downloading from {url}...")
    request = requests.get(url)
    with zipfile.ZipFile(BytesIO(request.content)) as zf:
        zf.extractall(str(project_path.parent))
    project_path.parent.joinpath(project_name + "-mix-master").rename(project_path)
    shutil.copy(
        _get_data_folder().joinpath("brownie-config.yaml"),
        project_path.joinpath("brownie-config.yaml"),
    )
    _add_to_sys_path(project_path)
    return str(project_path)


def from_ethpm(uri: str) -> "TempProject":

    """
    Generates a TempProject from an ethPM package.
    """

    manifest = get_manifest(uri)
    compiler_config = {
        "evm_version": None,
        "minify_source": False,
        "solc": {"version": None, "optimize": True, "runs": 200},
    }
    project = TempProject(manifest["package_name"], manifest["sources"], compiler_config)
    if web3.isConnected():
        for contract_name in project.keys():
            for address in get_deployment_addresses(manifest, contract_name):
                project[contract_name].at(address)
    return project


def _new_checks(project_path: Union[Path, str], ignore_subfolder: bool) -> Path:
    project_path = Path(project_path).resolve()
    if str(BROWNIE_FOLDER) in str(project_path):
        raise SystemError("Cannot make a new project inside the main brownie installation folder.")
    if not ignore_subfolder:
        check = check_for_project(project_path)
        if check and check != project_path:
            raise SystemError("Cannot make a new project in a subfolder of an existing project.")
    return project_path


def compile_source(
    source: str,
    solc_version: Optional[str] = None,
    optimize: bool = True,
    runs: Optional[int] = 200,
    evm_version: Optional[str] = None,
) -> "TempProject":
    """Compiles the given source code string and returns a TempProject container with
    the ContractContainer instances."""

    compiler_config: Dict = {"evm_version": evm_version, "minify_source": False}

    if solc_version is not None or source.lstrip().startswith("pragma"):
        compiler_config["solc"] = {"version": solc_version, "optimize": optimize, "runs": runs}
        return TempProject("TempSolcProject", {"<stdin>.sol": source}, compiler_config)

    return TempProject("TempVyperProject", {"<stdin>.vy": source}, compiler_config)


def load(project_path: Union[Path, str, None] = None, name: Optional[str] = None) -> "Project":
    """Loads a project and instantiates various related objects.

    Args:
        project_path: Path of the project to load. If None, will attempt to
                      locate a project using check_for_project()
        name: Name to assign to the project. If None, the name is generated
              from the name of the project folder

    Returns a Project object.
    """
    # checks
    if project_path is None:
        project_path = check_for_project(".")
    if not project_path or not _get_project_config_path(Path(project_path)):
        raise ProjectNotFound("Could not find Brownie project")

    project_path = Path(project_path).resolve()
    if name is None:
        name = project_path.name
        if not name.lower().endswith("project"):
            name += " project"
        name = "".join(i for i in name.title() if i.isalpha())
    if next((True for i in _loaded_projects if i._name == name), False):
        raise ProjectAlreadyLoaded("There is already a project loaded with this name")

    # paths
    _create_folders(project_path)
    _add_to_sys_path(project_path)

    # load sources and build
    return Project(name, project_path)


def _create_folders(project_path: Path) -> None:
    for path in [i for i in FOLDERS]:
        project_path.joinpath(path).mkdir(exist_ok=True)


def _add_to_sys_path(project_path: Path) -> None:
    project_path_string = str(project_path)
    if project_path_string in sys.path:
        return
    sys.path.insert(0, project_path_string)


def _compare_settings(left: Dict, right: Dict) -> bool:
    return next(
        (True for k, v in left.items() if v and not isinstance(v, dict) and v != right[k]), False
    )



#!/usr/bin/python3


from .account import Accounts
from .main import connect, disconnect, gas_limit, gas_price, is_connected, show_active  # NOQA 401
from .rpc import Rpc
from .state import TxHistory
from .web3 import web3

__all__ = ["accounts", "history", "rpc", "web3"]
__console_dir__ = ["connect", "disconnect", "show_active", "is_connected", "gas_limit", "gas_price"]

accounts = Accounts()
rpc = Rpc()
history = TxHistory()



#!/usr/bin/python3

import json
import sys
from typing import Any, Type

import psutil

# network


class UnknownAccount(Exception):
    pass


class UndeployedLibrary(Exception):
    pass


class UnsetENSName(Exception):
    pass


class IncompatibleEVMVersion(Exception):
    pass


class _RPCBaseException(Exception):
    def __init__(self, msg: str, cmd: str, proc: Type[psutil.Popen], uri: str) -> None:
        msg = f"{msg}\n\nCommand: {cmd}\nURI: {uri}\nExit Code: {proc.poll()}"
        if sys.platform != "win32":
            out = proc.stdout.read().decode().strip() or "  (Empty)"
            err = proc.stderr.read().decode().strip() or "  (Empty)"
            msg += f"\n\nStdout:\n{out}\n\nStderr:\n{err}"
        super().__init__(msg)


class RPCProcessError(_RPCBaseException):
    def __init__(self, cmd: str, proc: Type[psutil.Popen], uri: str) -> None:
        super().__init__("Unable to launch local RPC client.", cmd, proc, uri)


class RPCConnectionError(_RPCBaseException):
    def __init__(self, cmd: str, proc: Type[psutil.Popen], uri: str) -> None:
        super().__init__("Able to launch RPC client, but unable to connect.", cmd, proc, uri)


class RPCRequestError(Exception):
    pass


class MainnetUndefined(Exception):
    pass


class VirtualMachineError(Exception):

    """Raised when a call to a contract causes an EVM exception.

    Attributes:
        revert_msg: The returned error string, if any.
        source: The contract source code where the revert occured, if available."""

    def __init__(self, exc: Any) -> None:
        if type(exc) is not dict:
            try:
                exc = eval(str(exc))
            except SyntaxError:
                exc = {"message": str(exc)}
        self.revert_msg = msg = exc["message"]
        self.source = exc.get("source", "")
        if self.source:
            msg = f"{msg}\n{self.source}"
        super().__init__(msg)


class EventLookupError(LookupError):
    pass


class NamespaceCollision(AttributeError):
    pass


# project/


class ContractExists(Exception):
    pass


class ContractNotFound(Exception):
    pass


class ProjectAlreadyLoaded(Exception):
    pass


class ProjectNotFound(Exception):
    pass


class CompilerError(Exception):
    def __init__(self, e: Type[psutil.Popen]) -> None:
        err = [i["formattedMessage"] for i in json.loads(e.stdout_data)["errors"]]
        super().__init__("Compiler returned the following errors:\n\n" + "\n".join(err))


class IncompatibleSolcVersion(Exception):
    pass


class PragmaError(Exception):
    pass


class InvalidManifest(Exception):
    pass


class UnsupportedLanguage(Exception):
    pass



#!/usr/bin/python3

import json
from pathlib import Path

from brownie.network.state import TxHistory
from brownie.utils import color

COVERAGE_COLORS = [(0.8, "bright red"), (0.9, "bright yellow"), (1, "bright green")]


def _save_coverage_report(build, coverage_eval, report_path):
    # Saves a test coverage report for viewing in the GUI
    report = {
        "highlights": _get_highlights(build, coverage_eval),
        "coverage": _get_totals(build, coverage_eval),
        "sha1": {},  # TODO
    }
    report = json.loads(json.dumps(report, default=sorted))
    report_path = Path(report_path).absolute()
    if report_path.is_dir():
        report_path = report_path.joinpath("coverage.json")
    with report_path.open("w") as fp:
        json.dump(report, fp, sort_keys=True, indent=2)
    print(f"\nCoverage report saved at {report_path}")
    return report_path


def _print_gas_profile():
    # Formats and prints a gas profile report to the console
    print("\n\nGas Profile:")
    gas = TxHistory().gas_profile
    for i in sorted(gas):
        print(f"{i} -  avg: {gas[i]['avg']:.0f}  low: {gas[i]['low']}  high: {gas[i]['high']}")


def _print_coverage_totals(build, coverage_eval):
    # Formats and prints a coverage evaluation report to the console
    totals = _get_totals(build, coverage_eval)
    print("\n\nCoverage analysis:")
    for name in sorted(totals):
        pct = _pct(totals[name]["totals"]["statements"], totals[name]["totals"]["branches"])
        print(f"\n  contract: {color['contract']}{name}{color} - {_cov_color(pct)}{pct:.1%}{color}")
        cov = totals[name]
        for fn_name, count in cov["statements"].items():
            branch = cov["branches"][fn_name] if fn_name in cov["branches"] else (0, 0, 0)
            pct = _pct(count, branch)
            print(f"    {fn_name} - {_cov_color(pct)}{pct:.1%}{color}")


def _cov_color(pct):
    return color(next(i[1] for i in COVERAGE_COLORS if pct <= i[0]))


def _pct(statement, branch):
    pct = statement[0] / statement[1]
    if branch[-1]:
        pct = (pct + (branch[0] + branch[1]) / (branch[2] * 2)) / 2
    return pct


def _get_totals(build, coverage_eval):
    # Returns a modified coverage eval dict showing counts and totals for each function.

    coverage_eval = _split_by_fn(build, coverage_eval)
    results = dict(
        (
            i,
            {
                "statements": {},
                "totals": {"statements": 0, "branches": [0, 0]},
                "branches": {"true": {}, "false": {}},
            },
        )
        for i in coverage_eval
    )
    for name in coverage_eval:
        coverage_map = build.get(name)["coverageMap"]
        r = results[name]
        r["statements"], r["totals"]["statements"] = _statement_totals(
            coverage_eval[name], coverage_map["statements"]
        )
        r["branches"], r["totals"]["branches"] = _branch_totals(
            coverage_eval[name], coverage_map["branches"]
        )
    return results


def _split_by_fn(build, coverage_eval):
    # Splits a coverage eval dict so that coverage indexes are stored by function.
    results = dict(
        (i, {"statements": {}, "branches": {"true": {}, "false": {}}}) for i in coverage_eval
    )
    for name in coverage_eval:
        map_ = build.get(name)["coverageMap"]
        results[name] = dict((k, _split(v, map_, k)) for k, v in coverage_eval[name].items())
    return results


def _split(coverage_eval, coverage_map, key):
    results = {}
    for fn, map_ in coverage_map["statements"][key].items():
        results[fn] = [[i for i in map_ if int(i) in coverage_eval[0]], [], []]
    for fn, map_ in coverage_map["branches"][key].items():
        results[fn][1] = [i for i in map_ if int(i) in coverage_eval[1]]
        results[fn][2] = [i for i in map_ if int(i) in coverage_eval[2]]
    return results


def _statement_totals(coverage_eval, coverage_map):
    result = {}
    count, total = 0, 0
    for path, fn in [(k, x) for k, v in coverage_eval.items() for x in v]:
        count += len(coverage_eval[path][fn][0])
        total += len(coverage_map[path][fn])
        result[fn] = (len(coverage_eval[path][fn][0]), len(coverage_map[path][fn]))
    return result, (count, total)


def _branch_totals(coverage_eval, coverage_map):
    result = {}
    final = [0, 0, 0]
    for path, fn in [(k, x) for k, v in coverage_map.items() for x in v]:
        if path not in coverage_eval:
            true, false = 0, 0
        else:
            true = len(coverage_eval[path][fn][2])
            false = len(coverage_eval[path][fn][1])
        total = len(coverage_map[path][fn])
        result[fn] = (true, false, total)
        for i in range(3):
            final[i] += result[fn][i]
    return result, final


def _get_highlights(build, coverage_eval):
    # Returns a highlight map formatted for display in the GUI.
    results = {"statements": {}, "branches": {}}
    for name, eval_ in coverage_eval.items():
        coverage_map = build.get(name)["coverageMap"]
        results["statements"][name] = _statement_highlights(eval_, coverage_map["statements"])
        results["branches"][name] = _branch_highlights(eval_, coverage_map["branches"])
    return results


def _statement_highlights(coverage_eval, coverage_map):
    results = dict((i, []) for i in coverage_map)
    for path, fn in [(k, x) for k, v in coverage_map.items() for x in v]:
        results[path].extend(
            [
                list(offset) + [_statement_color(i, coverage_eval, path), ""]
                for i, offset in coverage_map[path][fn].items()
            ]
        )
    return results


def _statement_color(i, coverage_eval, path):
    if path not in coverage_eval or int(i) not in coverage_eval[path][0]:
        return "red"
    return "green"


def _branch_highlights(coverage_eval, coverage_map):
    results = dict((i, []) for i in coverage_map)
    for path, fn in [(k, x) for k, v in coverage_map.items() for x in v]:
        results[path].extend(
            [
                list(offset[:2]) + [_branch_color(int(i), coverage_eval, path, offset[2]), ""]
                for i, offset in coverage_map[path][fn].items()
            ]
        )
    return results


def _branch_color(i, coverage_eval, path, jump):
    if path not in coverage_eval:
        return "red"
    if i in coverage_eval[path][2]:
        if i in coverage_eval[path][1]:
            return "green"
        return "yellow" if jump else "orange"
    if i in coverage_eval[path][1]:
        return "orange" if jump else "yellow"
    return "red"



#!/usr/bin/python3

from collections import deque
from hashlib import sha1
from typing import Dict, List, Optional, Tuple

import vyper
from semantic_version import Version
from vyper.cli import vyper_json

from brownie.project.compiler.utils import expand_source_map
from brownie.project.sources import is_inside_offset


def get_version() -> Version:
    return Version.coerce(vyper.__version__)


def compile_from_input_json(
    input_json: Dict, silent: bool = True, allow_paths: Optional[str] = None
) -> Dict:

    """
    Compiles contracts from a standard input json.

    Args:
        input_json: solc input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    if not silent:
        print("Compiling contracts...")
        print(f"  Vyper version: {get_version()}")
    return vyper_json.compile_json(input_json, root_path=allow_paths)


def _get_unique_build_json(
    output_evm: Dict, path_str: str, contract_name: str, ast_json: List, offset: Tuple
) -> Dict:
    pc_map, statement_map, branch_map = _generate_coverage_data(
        output_evm["deployedBytecode"]["sourceMap"],
        output_evm["deployedBytecode"]["opcodes"],
        path_str,
        contract_name,
        ast_json,
    )
    return {
        "allSourcePaths": [path_str],
        "bytecode": output_evm["bytecode"]["object"],
        "bytecodeSha1": sha1(output_evm["bytecode"]["object"].encode()).hexdigest(),
        "coverageMap": {"statements": statement_map, "branches": branch_map},
        "dependencies": _get_dependencies(ast_json),
        "offset": offset,
        "pcMap": pc_map,
        "type": "contract",
    }


def _get_dependencies(ast_json: List) -> List:
    import_nodes = [i for i in ast_json if i["ast_type"] == "Import"]
    import_nodes += [
        i for i in ast_json if i["ast_type"] == "ImportFrom" if i["module"] != "vyper.interfaces"
    ]
    return sorted(set([i["names"][0]["name"].split(".")[-1] for i in import_nodes]))


def _generate_coverage_data(
    source_map_str: str, opcodes_str: str, source_str: str, contract_name: str, ast_json: List
) -> Tuple:
    if not opcodes_str:
        return {}, {}, {}

    source_map = deque(expand_source_map(source_map_str))
    opcodes = deque(opcodes_str.split(" "))

    fn_nodes = [i for i in ast_json if i["ast_type"] == "FunctionDef"]
    fn_offsets = dict((i["name"], _convert_src(i["src"])) for i in fn_nodes)
    stmt_nodes = set(_convert_src(i["src"]) for i in _get_statement_nodes(fn_nodes))

    statement_map: Dict = {}
    branch_map: Dict = {}

    pc_list: List = []
    count, pc = 0, 0

    while opcodes:
        # format of source is [start, stop, contract_id, jump code]
        source = source_map.popleft()
        pc_list.append({"op": opcodes.popleft(), "pc": pc})

        if source[3] != "-":
            pc_list[-1]["jump"] = source[3]

        pc += 1
        if opcodes and opcodes[0][:2] == "0x":
            pc_list[-1]["value"] = opcodes.popleft()
            pc += int(pc_list[-1]["op"][4:])

        # set source offset (-1 means none)
        if source[0] == -1:
            continue
        offset = (source[0], source[0] + source[1])
        pc_list[-1]["path"] = source_str
        pc_list[-1]["offset"] = offset

        try:
            if "offset" in pc_list[-2] and offset == pc_list[-2]["offset"]:
                pc_list[-1]["fn"] = pc_list[-2]["fn"]
            else:
                # statement coverage
                fn = next(k for k, v in fn_offsets.items() if is_inside_offset(offset, v))
                pc_list[-1]["fn"] = f"{contract_name}.{fn}"
                stmt_offset = next(i for i in stmt_nodes if is_inside_offset(offset, i))
                stmt_nodes.remove(stmt_offset)
                statement_map.setdefault(pc_list[-1]["fn"], {})[count] = stmt_offset
                pc_list[-1]["statement"] = count
                count += 1
        except (KeyError, IndexError, StopIteration):
            pass

        if pc_list[-1]["op"] not in ("JUMPI", "REVERT"):
            continue

        node = _find_node_by_offset(ast_json, offset)
        if pc_list[-1]["op"] == "REVERT":
            # custom revert error strings
            if node["ast_type"] == "FunctionDef" and pc_list[-7]["op"] == "CALLVALUE":
                pc_list[-1]["dev"] = "Cannot send ether to nonpayable function"
            elif node["ast_type"] == "Subscript":
                pc_list[-1]["dev"] = "Index out of range"
            elif node["ast_type"] in ("AugAssign", "BinOp"):
                if node["op"]["ast_type"] == "Sub":
                    pc_list[-1]["dev"] = "Integer underflow"
                elif node["op"]["ast_type"] == "Div":
                    pc_list[-1]["dev"] = "Division by zero"
                elif node["op"]["ast_type"] == "Mod":
                    pc_list[-1]["dev"] = "Modulo by zero"
                else:
                    pc_list[-1]["dev"] = "Integer overflow"
            continue

        if node["ast_type"] in ("Assert", "If") or (
            node["ast_type"] == "Expr"
            and node["value"]["func"].get("id", None) == "assert_modifiable"
        ):
            # branch coverage
            pc_list[-1]["branch"] = count
            branch_map.setdefault(pc_list[-1]["fn"], {})
            if node["ast_type"] == "If":
                branch_map[pc_list[-1]["fn"]][count] = _convert_src(node["test"]["src"]) + (False,)
            else:
                branch_map[pc_list[-1]["fn"]][count] = offset + (True,)
            count += 1

    pc_list[0]["path"] = source_str
    pc_list[0]["offset"] = [0, _convert_src(ast_json[-1]["src"])[1]]
    pc_map = dict((i.pop("pc"), i) for i in pc_list)

    return pc_map, {source_str: statement_map}, {source_str: branch_map}


def _convert_src(src: str) -> Tuple[int, int]:
    if src is None:
        return -1, -1
    src_int = [int(i) for i in src.split(":")[:2]]
    return src_int[0], src_int[0] + src_int[1]


def _find_node_by_offset(ast_json: List, offset: Tuple) -> Dict:
    node = next(i for i in ast_json if is_inside_offset(offset, _convert_src(i["src"])))
    if _convert_src(node["src"]) == offset:
        return node
    node_list = [i for i in node.values() if isinstance(i, dict) and "ast_type" in i]
    node_list.extend([x for i in node.values() if isinstance(i, list) for x in i])
    return _find_node_by_offset(node_list, offset)


def _get_statement_nodes(ast_json: List) -> List:
    stmt_nodes: List = []
    for node in ast_json:
        children = [x for v in node.values() if isinstance(v, list) for x in v]
        if children:
            stmt_nodes += _get_statement_nodes(children)
        else:
            stmt_nodes.append(node)
    return stmt_nodes



#!/usr/bin/python3

import json
from hashlib import sha1
from pathlib import Path

from brownie._config import ARGV
from brownie.network.state import _get_current_dependencies
from brownie.project.scripts import _get_ast_hash
from brownie.test import coverage

STATUS_SYMBOLS = {"passed": ".", "skipped": "s", "failed": "F"}

STATUS_TYPES = {
    ".": "passed",
    "s": "skipped",
    "F": "failed",
    "E": "error",
    "x": "xfailed",
    "X": "xpassed",
}


class TestManager:
    def __init__(self, project):
        self.project = project
        self.project_path = project._path
        self.active_path = None
        self.count = 0
        self.results = None
        self.isolated = set()
        glob = self.project_path.glob("tests/**/conftest.py")
        self.conf_hashes = dict((self._path(i.parent), _get_ast_hash(i)) for i in glob)
        try:
            with self.project_path.joinpath("build/tests.json").open() as fp:
                hashes = json.load(fp)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            hashes = {"tests": {}, "contracts": {}, "tx": {}}

        self.tests = dict(
            (k, v)
            for k, v in hashes["tests"].items()
            if Path(k).exists() and self._get_hash(k) == v["sha1"]
        )
        build = self.project._build
        self.contracts = dict((k, v["bytecodeSha1"]) for k, v in build.items() if v["bytecode"])
        changed_contracts = set(
            k
            for k, v in hashes["contracts"].items()
            if k not in self.contracts or v != self.contracts[k]
        )
        if changed_contracts:
            for txhash, coverage_eval in hashes["tx"].items():
                if not changed_contracts.intersection(coverage_eval.keys()):
                    coverage._add_cached_transaction(txhash, coverage_eval)
            self.tests = dict(
                (k, v)
                for k, v in self.tests.items()
                if v["isolated"] is not False and not changed_contracts.intersection(v["isolated"])
            )
        else:
            for txhash, coverage_eval in hashes["tx"].items():
                coverage._add_cached_transaction(txhash, coverage_eval)

    def _path(self, path):
        return str(Path(path).absolute().relative_to(self.project_path))

    def set_isolated_modules(self, paths):
        self.isolated = set(self._path(i) for i in paths)

    def _get_hash(self, path):
        hash_ = _get_ast_hash(path)
        for confpath in filter(lambda k: k in path, sorted(self.conf_hashes)):
            hash_ += self.conf_hashes[confpath]
        return sha1(hash_.encode()).hexdigest()

    def check_updated(self, path):
        path = self._path(path)
        if path not in self.tests or not self.tests[path]["isolated"]:
            return False
        if ARGV["coverage"] and not self.tests[path]["coverage"]:
            return False
        for txhash in self.tests[path]["txhash"]:
            coverage._check_cached(txhash, False)
        return True

    def module_completed(self, path):
        path = self._path(path)
        isolated = False
        if path in self.isolated:
            isolated = [i for i in _get_current_dependencies() if i in self.contracts]
        txhash = coverage._get_active_txlist()
        coverage._clear_active_txlist()
        if not ARGV["coverage"] and (path in self.tests and self.tests[path]["coverage"]):
            txhash = self.tests[path]["txhash"]
        self.tests[path] = {
            "sha1": self._get_hash(path),
            "isolated": isolated,
            "coverage": ARGV["coverage"] or (path in self.tests and self.tests[path]["coverage"]),
            "txhash": txhash,
            "results": "".join(self.results),
        }

    def save_json(self):
        txhash = set(x for v in self.tests.values() for x in v["txhash"])
        coverage_eval = dict((k, v) for k, v in coverage.get_coverage_eval().items() if k in txhash)
        report = {"tests": self.tests, "contracts": self.contracts, "tx": coverage_eval}
        with self.project_path.joinpath("build/tests.json").open("w") as fp:
            json.dump(report, fp, indent=2, sort_keys=True, default=sorted)

    def set_active(self, path):
        path = self._path(path)
        if path == self.active_path:
            self.count += 1
            return
        self.active_path = path
        self.count = 0
        if path in self.tests and ARGV["update"]:
            self.results = list(self.tests[path]["results"])
        else:
            self.results = []

    def check_status(self, report):
        if report.when == "setup":
            self._skip = report.skipped
            if len(self.results) < self.count + 1:
                self.results.append("s" if report.skipped else None)
            if report.failed:
                self.results[self.count] = "E"
                return "error", "E", "ERROR"
            return "", "", ""
        if report.when == "teardown":
            if report.failed:
                self.results[self.count] = "E"
                return "error", "E", "ERROR"
            elif self._skip:
                report.outcome = STATUS_TYPES[self.results[self.count]]
                return "skipped", "s", "SKIPPED"
            return "", "", ""
        if hasattr(report, "wasxfail"):
            self.results[self.count] = "x" if report.skipped else "X"
            if report.skipped:
                return "xfailed", "x", "XFAIL"
            elif report.passed:
                return "xpassed", "X", "XPASS"
        self.results[self.count] = STATUS_SYMBOLS[report.outcome]
        return report.outcome, STATUS_SYMBOLS[report.outcome], report.outcome.upper()



#!/usr/bin/python3

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    requirements = list(map(str.strip, f.read().split("\n")))[:-1]

setup(
    name="eth-brownie",
    packages=find_packages(),
    version="1.4.1",  # don't change this manually, use bumpversion instead
    license="MIT",
    description="A Python framework for Ethereum smart contract deployment, testing and interaction.",  # noqa: E501
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ben Hauser",
    author_email="ben@hauser.id",
    url="https://github.com/iamdefinitelyahuman/brownie",
    keywords=["brownie"],
    install_requires=requirements,
    entry_points={
        "console_scripts": ["brownie=brownie._cli.__main__:main"],
        "pytest11": ["pytest-brownie=brownie.test.plugin"],
    },
    include_package_data=True,
    python_requires=">=3.6,<4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



#!/usr/bin/python3

from typing import Any, Dict, Iterator, List

from brownie._singleton import _Singleton

from .rpc import _revert_register
from .web3 import _resolve_address

_contract_map: Dict = {}


class TxHistory(metaclass=_Singleton):

    """List-like singleton container that contains TransactionReceipt objects.
    Whenever a transaction is broadcast, the TransactionReceipt is automatically
    added to this container."""

    def __init__(self) -> None:
        self._list: List = []
        self.gas_profile: Dict = {}
        _revert_register(self)

    def __repr__(self) -> str:
        return str(self._list)

    def __bool__(self) -> bool:
        return bool(self._list)

    def __contains__(self, item: Any) -> bool:
        return item in self._list

    def __iter__(self) -> Iterator:
        return iter(self._list)

    def __getitem__(self, key: Any) -> Any:
        return self._list[key]

    def __len__(self) -> int:
        return len(self._list)

    def _reset(self) -> None:
        self._list.clear()

    def _revert(self, height: int) -> None:
        self._list = [i for i in self._list if i.block_number <= height]

    def _add_tx(self, tx: Any) -> None:
        self._list.append(tx)

    def clear(self) -> None:
        self._list.clear()

    def copy(self) -> List:
        """Returns a shallow copy of the object as a list"""
        return self._list.copy()

    def from_sender(self, account: str) -> List:
        """Returns a list of transactions where the sender is account"""
        return [i for i in self._list if i.sender == account]

    def to_receiver(self, account: str) -> List:
        """Returns a list of transactions where the receiver is account"""
        return [i for i in self._list if i.receiver == account]

    def of_address(self, account: str) -> List:
        """Returns a list of transactions where account is the sender or receiver"""
        return [i for i in self._list if i.receiver == account or i.sender == account]

    def _gas(self, fn_name: str, gas_used: int) -> None:
        if fn_name not in self.gas_profile:
            self.gas_profile[fn_name] = {
                "avg": gas_used,
                "high": gas_used,
                "low": gas_used,
                "count": 1,
            }
            return
        gas = self.gas_profile[fn_name]
        gas.update(
            {
                "avg": (gas["avg"] * gas["count"] + gas_used) // (gas["count"] + 1),
                "high": max(gas["high"], gas_used),
                "low": min(gas["low"], gas_used),
            }
        )
        gas["count"] += 1


def _find_contract(address: Any) -> Any:
    address = _resolve_address(address)
    if address not in _contract_map:
        return None
    return _contract_map[address]


def _get_current_dependencies() -> List:
    dependencies = set(v._name for v in _contract_map.values())
    for contract in _contract_map.values():
        dependencies.update(contract._build["dependencies"])
    return sorted(dependencies)


def _add_contract(contract: Any) -> None:
    _contract_map[contract.address] = contract


def _remove_contract(contract: Any) -> None:
    del _contract_map[contract.address]



#!/usr/bin/python3

import importlib
import json
import re
import time
from os import environ
from pathlib import Path

from docopt import docopt
from pythx import Client
from pythx.middleware.toolname import ClientToolNameMiddleware

from brownie import project
from brownie._cli.__main__ import __version__
from brownie._config import ARGV, _update_argv_from_docopt
from brownie.exceptions import ProjectNotFound
from brownie.utils import color, notify

__doc__ = f"""Usage: brownie analyze [options] [--async | --interval=<sec>]

Options:
  --gui                     Launch the Brownie GUI after analysis
  --full                    Perform a full scan (MythX Pro required)
  --interval=<sec>          Result polling interval in seconds [default: 3]
  --async                   Do not poll for results, print job IDs and exit
  --access-token=<string>   The JWT access token from the MythX dashboard
  --eth-address=<string>    The address of your MythX account
  --password=<string>       The password of your MythX account
  --help -h                 Display this message

Use the "analyze" command to submit your project to the MythX API for
smart contract security analysis.


To authenticate with the MythX API, it is recommended that you provide
the MythX JWT access token. It can be obtained on the MythX dashboard
site in the profile section. They should be passed through the environment
variable "MYTHX_ACCESS_TOKEN". If that is not possible, it can also be
passed explicitly with the respective command line option.

Alternatively, you have to provide a username/password combination. It
is recommended to pass them through the environment variables as
"MYTHX_ETH_ADDRESS" and "MYTHX_PASSWORD".

You can also choose to not authenticate and submit your analyses as a free
trial user. No registration required! To see your past analyses, get access
to deeper vulnerability detection, and a neat dashboard, register at
https://mythx.io/. Any questions? Hit up dominik.muhs@consensys.net or contact
us on the website!
"""


SEVERITY_COLOURS = {"LOW": "green", "MEDIUM": "yellow", "HIGH": "red"}
DASHBOARD_BASE_URL = "https://dashboard.mythx.io/#/console/analyses/"
TRIAL_PRINTED = False
BYTECODE_ADDRESS_PATCH = re.compile(r"__\w{38}")
DEPLOYED_ADDRESS_PATCH = re.compile(r"__\$\w{34}\$__")


def construct_source_dict_from_artifact(artifact):
    return {
        artifact.get("sourcePath"): {
            "source": artifact.get("source"),
            # "ast": artifact.get("ast"),  # NOTE: Reenable once container issue fixed
        }
    }


def construct_request_from_artifact(artifact):
    global BYTECODE_ADDRESS_PATCH

    bytecode = artifact.get("bytecode")
    deployed_bytecode = artifact.get("deployedBytecode")
    source_map = artifact.get("sourceMap")
    deployed_source_map = artifact.get("deployedSourceMap")

    bytecode = re.sub(BYTECODE_ADDRESS_PATCH, "0" * 40, bytecode)
    deployed_bytecode = re.sub(DEPLOYED_ADDRESS_PATCH, "0" * 40, deployed_bytecode)

    source_list = artifact.get("allSourcePaths")
    return {
        "contract_name": artifact.get("contractName"),
        "bytecode": bytecode if bytecode else None,
        "deployed_bytecode": deployed_bytecode if deployed_bytecode else None,
        "source_map": source_map if source_map else None,
        "deployed_source_map": deployed_source_map if deployed_source_map else None,
        "sources": construct_source_dict_from_artifact(artifact),
        "source_list": source_list if source_list else None,
        "main_source": artifact.get("sourcePath"),
        "solc_version": artifact["compiler"]["version"],
        "analysis_mode": "full" if ARGV["full"] else "quick",
    }


def get_mythx_client():
    # if both eth address and username are None,
    if ARGV["access-token"]:
        authenticated = True
        auth_args = {"access_token": ARGV["access-token"]}
    elif ARGV["eth-address"] and ARGV["password"]:
        authenticated = True
        auth_args = {"eth_address": ARGV["eth-address"], "password": ARGV["password"]}
    elif environ.get("MYTHX_ACCESS_TOKEN"):
        authenticated = True
        auth_args = {"access_token": environ.get("MYTHX_ACCESS_TOKEN")}
    elif environ.get("MYTHX_ETH_ADDRESS") and environ.get("MYTHX_PASSWORD"):
        authenticated = True
        auth_args = {
            "eth_address": environ.get("MYTHX_ETH_ADDRESS"),
            "password": environ.get("MYTHX_PASSWORD"),
        }
    else:
        authenticated = False
        auth_args = {
            "eth_address": "0x0000000000000000000000000000000000000000",
            "password": "trial",
        }

    return (
        Client(
            **auth_args,
            middlewares=[ClientToolNameMiddleware(name="brownie-{}".format(__version__))],
        ),
        authenticated,
    )


def get_contract_locations(build):
    return {d["sourcePath"]: d["contractName"] for _, d in build.items()}


def get_contract_types(build):
    contracts = set([n for n, d in build.items() if d["type"] == "contract"])
    libraries = set([n for n, d in build.items() if d["type"] == "library"])

    return contracts, libraries


def assemble_contract_jobs(build, contracts):
    job_data = {}
    for contract in contracts:
        artifact = build.get(contract)
        job_data[contract] = construct_request_from_artifact(artifact)
    return job_data


def update_contract_jobs_with_dependencies(build, contracts, libraries, job_data):
    for lib in libraries:
        artifact = build.get(lib)
        source_dict = construct_source_dict_from_artifact(artifact)
        deps = set(build.get_dependents(lib))
        dep_contracts = contracts.intersection(deps)
        for contract in dep_contracts:
            job_data[contract]["sources"].update(source_dict)

    return job_data


def send_to_mythx(job_data, client, authenticated):
    job_uuids = []
    for c, (contract_name, analysis_request) in enumerate(job_data.items(), start=1):
        resp = client.analyze(**analysis_request)
        if ARGV["async"] and authenticated:
            print(
                "The analysis for {} can be found at {}{}".format(
                    contract_name, DASHBOARD_BASE_URL, resp.uuid
                )
            )
        else:
            print(
                f"Submitted analysis {color['value']}{resp.uuid}{color} for "
                f"contract {color['contract']}{contract_name}{color} ({c}/{len(job_data)})"
            )
        job_uuids.append(resp.uuid)

    return job_uuids


def wait_for_jobs(job_uuids, client):
    for uuid in job_uuids:
        while not client.analysis_ready(uuid):
            time.sleep(int(ARGV["interval"]))


def print_trial_message(issue):
    global TRIAL_PRINTED
    if issue.swc_id == "" or issue.severity.name in ("UNKNOWN", "NONE"):
        if not TRIAL_PRINTED:
            print(f"\n{issue.description_short}")
            print(f"{issue.description_long}\n")
            TRIAL_PRINTED = True
        return True
    return False


def update_report(client, uuid, highlight_report, stdout_report, source_to_name):
    resp = client.report(uuid)
    for report in resp.issue_reports:
        for issue in resp:
            if print_trial_message(issue):
                continue

            # convert issue locations to report locations
            # severities are highlighted according to SEVERITY_COLOURS
            for loc in issue.locations:
                comp = loc.source_map.components[0]
                source_list = loc.source_list or report.source_list

                if source_list and 0 <= comp.file_id < len(source_list):
                    filename = source_list[comp.file_id]
                    if filename not in source_to_name:
                        continue
                    contract_name = source_to_name[filename]
                    severity = issue.severity.name
                    stdout_report.setdefault(contract_name, {}).setdefault(severity, 0)
                    stdout_report[contract_name][severity] += 1
                    highlight_report["highlights"]["MythX"].setdefault(
                        contract_name, {filename: []}
                    )
                    highlight_report["highlights"]["MythX"][contract_name][filename].append(
                        [
                            comp.offset,
                            comp.offset + comp.length,
                            SEVERITY_COLOURS[severity],
                            f"{issue.swc_id}: {issue.description_short}\n{issue.description_long}",
                        ]
                    )


def main():
    args = docopt(__doc__)
    _update_argv_from_docopt(args)

    project_path = project.check_for_project(".")
    if project_path is None:
        raise ProjectNotFound

    build = project.load()._build

    print("Preparing project data for submission to MythX...")
    contracts, libraries = get_contract_types(build)

    job_data = assemble_contract_jobs(build, contracts)
    job_data = update_contract_jobs_with_dependencies(build, contracts, libraries, job_data)

    client, authenticated = get_mythx_client()

    job_uuids = send_to_mythx(job_data, client, authenticated)

    # exit if user wants an async analysis run
    if ARGV["async"] and authenticated:
        print(
            "\nAll contracts were submitted successfully. Check the dashboard at "
            "https://dashboard.mythx.io/ for the progress and results of your analyses"
        )
        return

    print("\nWaiting for results...")
    wait_for_jobs(job_uuids, client)

    # assemble report json
    source_to_name = get_contract_locations(build)
    highlight_report = {"highlights": {"MythX": {}}}
    stdout_report = {}
    for c, uuid in enumerate(job_uuids, start=1):
        print(f"Generating report for job {color['value']}{uuid}{color} ({c}/{len(job_uuids)})")
        if authenticated:
            print("You can also check the results at {}{}\n".format(DASHBOARD_BASE_URL, uuid))

        update_report(client, uuid, highlight_report, stdout_report, source_to_name)

    # erase previous report
    report_path = Path("reports/security.json")
    if report_path.exists():
        report_path.unlink()

    total_issues = sum(x for i in stdout_report.values() for x in i.values())
    if not total_issues:
        notify("SUCCESS", "No issues found!")
        return

    # display console report
    total_high_severity = sum(i.get("HIGH", 0) for i in stdout_report.values())
    if total_high_severity:
        notify(
            "WARNING", f"Found {total_issues} issues including {total_high_severity} high severity!"
        )
    else:
        print(f"Found {total_issues} issues:")
    for name in sorted(stdout_report):
        print(f"\n  contract: {color['contract']}{name}{color}")
        for key in [i for i in ("HIGH", "MEDIUM", "LOW") if i in stdout_report[name]]:
            c = color("bright " + SEVERITY_COLOURS[key])
            print(f"    {key.title()}: {c}{stdout_report[name][key]}{color}")

    # Write report to Brownie directory
    with report_path.open("w+") as fp:
        json.dump(highlight_report, fp, indent=2, sort_keys=True)

    # Launch GUI if user requested it
    if ARGV["gui"]:
        print("Launching the Brownie GUI")
        Gui = importlib.import_module("brownie._gui").Gui
        Gui().mainloop()



#!/usr/bin/python3

import json
import shutil
import sys
from pathlib import Path

from docopt import docopt

from brownie import accounts
from brownie._config import _get_data_folder
from brownie.convert import to_address
from brownie.utils import color, notify

__doc__ = """Usage: brownie accounts <command> [<arguments> ...] [options]

Commands:
  list                             List available accounts
  new <id>                         Add a new account by entering a private key
  generate <id>                    Add a new account with a random private key
  import <id> <path>               Import a new account via a keystore file
  export <id> <path>               Export an existing account keystore file
  password <id>                    Change the password of an existing account
  delete <id>                      Delete an existing account

Options:
  --help -h                        Display this message

Command-line helper for managing local accounts. You can unlock local accounts from
scripts or the console using the Accounts.load method.
"""


def main():
    args = docopt(__doc__)
    try:
        fn = getattr(sys.modules[__name__], f"_{args['<command>']}")
    except AttributeError:
        print("Invalid command. Try brownie accounts --help")
        return
    try:
        fn(*args["<arguments>"])
    except TypeError:
        print(f"Invalid arguments for command '{args['<command>']}'. Try brownie ethpm --help")
        return


def _list():
    account_paths = sorted(_get_data_folder().glob("accounts/*.json"))
    print(f"Found {len(account_paths)} account{'s' if len(account_paths)!=1 else ''}:")
    for path in account_paths:
        u = "\u2514" if path == account_paths[-1] else "\u251c"
        with path.open() as fp:
            data = json.load(fp)
        print(
            f" {color('bright black')}{u}\u2500{color('bright blue')}{path.stem}{color}"
            f": {color('bright magenta')}{to_address(data['address'])}{color}"
        )


def _new(id_):
    pk = input("Enter the private key you wish to add: ")
    a = accounts.add(pk)
    a.save(id_)
    notify(
        "SUCCESS",
        f"A new account '{color('bright magenta')}{a.address}{color}'"
        f" has been generated with the id '{color('bright blue')}{id_}{color}'",
    )


def _generate(id_):
    print("Generating a new private key...")
    a = accounts.add()
    a.save(id_)
    notify(
        "SUCCESS",
        f"A new account '{color('bright magenta')}{a.address}{color}'"
        f" has been generated with the id '{color('bright blue')}{id_}{color}'",
    )


def _import(id_, path):
    source_path = Path(path).absolute()
    if not source_path.suffix:
        source_path = source_path.with_suffix(".json")
    dest_path = _get_data_folder().joinpath(f"accounts/{id_}.json")
    if dest_path.exists():
        raise FileExistsError(f"A keystore file already exists with the id '{id_}'")
    accounts.load(source_path)
    shutil.copy(source_path, dest_path)
    notify(
        "SUCCESS",
        f"Keystore '{color('bright magenta')}{source_path}{color}'"
        f" has been imported with the id '{color('bright blue')}{id_}{color}'",
    )


def _export(id_, path):
    source_path = _get_data_folder().joinpath(f"accounts/{id_}.json")
    if not source_path.exists():
        raise FileNotFoundError(f"No keystore exists with the id '{id_}'")
    dest_path = Path(path).absolute()
    if not dest_path.suffix:
        dest_path = dest_path.with_suffix(".json")
    if dest_path.exists():
        raise FileExistsError(f"Export path {dest_path} already exists")
    shutil.copy(source_path, dest_path)
    notify(
        "SUCCESS",
        f"Account with id '{color('bright blue')}{id_}{color}' has been"
        f" exported to keystore '{color('bright magenta')}{dest_path}{color}'",
    )


def _password(id_):
    a = accounts.load(id_)
    a.save(id_, overwrite=True)
    notify("SUCCESS", f"Password has been changed for account '{color('bright blue')}{id_}{color}'")


def _delete(id_):
    path = _get_data_folder().joinpath(f"accounts/{id_}.json")
    path.unlink()
    notify("SUCCESS", f"Account '{color('bright blue')}{id_}{color}' has been deleted")



#!/usr/bin/python3

from copy import deepcopy
from typing import Dict, Optional, Union

from semantic_version import Version

from brownie.exceptions import UnsupportedLanguage
from brownie.project import sources
from brownie.project.compiler.solidity import (  # NOQA: F401
    find_best_solc_version,
    find_solc_versions,
    install_solc,
    set_solc_version,
)

from . import solidity, vyper

STANDARD_JSON: Dict = {
    "language": None,
    "sources": {},
    "settings": {
        "outputSelection": {
            "*": {"*": ["abi", "evm.bytecode", "evm.deployedBytecode"], "": ["ast"]}
        },
        "evmVersion": None,
        "remappings": [],
    },
}
EVM_SOLC_VERSIONS = [
    # ("istanbul", Version("0.5.13")),  # TODO enable when ganache istanbul support is out of beta
    ("petersburg", Version("0.5.5")),
    ("byzantium", Version("0.4.0")),
]


def compile_and_format(
    contract_sources: Dict[str, str],
    solc_version: Optional[str] = None,
    optimize: bool = True,
    runs: int = 200,
    evm_version: int = None,
    minify: bool = False,
    silent: bool = True,
    allow_paths: Optional[str] = None,
) -> Dict:
    """Compiles contracts and returns build data.

    Args:
        contracts: a dictionary in the form of {'path': "source code"}
        solc_version: solc version to compile with (use None to set via pragmas)
        optimize: enable solc optimizer
        runs: optimizer runs
        evm_version: evm version to compile for
        minify: minify source files
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns:
        build data dict
    """
    if not contract_sources:
        return {}

    if [i for i in contract_sources if not i.endswith(".sol") and not i.endswith(".vy")]:
        raise UnsupportedLanguage("Source filenames must end in .sol or .vy")

    build_json: Dict = {}
    compiler_targets = {}

    vyper_paths = [i for i in contract_sources if i.endswith(".vy")]
    if vyper_paths:
        compiler_targets["vyper"] = vyper_paths

    solc_sources = dict((k, v) for k, v in contract_sources.items() if k.endswith(".sol"))
    if solc_sources:
        if solc_version is None:
            compiler_targets.update(
                find_solc_versions(solc_sources, install_needed=True, silent=silent)
            )
        else:
            compiler_targets[solc_version] = list(solc_sources)

    for version, path_list in compiler_targets.items():
        compiler_data: Dict = {"minify_source": minify}
        if version == "vyper":
            language = "Vyper"
            compiler_data["version"] = str(vyper.get_version())
        else:
            set_solc_version(version)
            language = "Solidity"
            compiler_data["version"] = str(solidity.get_version())

        to_compile = dict((k, v) for k, v in contract_sources.items() if k in path_list)

        input_json = generate_input_json(to_compile, optimize, runs, evm_version, minify, language)
        output_json = compile_from_input_json(input_json, silent, allow_paths)
        build_json.update(generate_build_json(input_json, output_json, compiler_data, silent))
    return build_json


def generate_input_json(
    contract_sources: Dict[str, str],
    optimize: bool = True,
    runs: int = 200,
    evm_version: Union[int, str, None] = None,
    minify: bool = False,
    language: str = "Solidity",
) -> Dict:

    """Formats contracts to the standard solc input json.

    Args:
        contract_sources: a dictionary in the form of {path: 'source code'}
        optimize: enable solc optimizer
        runs: optimizer runs
        evm_version: evm version to compile for
        minify: should source code be minified?
        language: source language (Solidity or Vyper)

    Returns: dict
    """

    if language not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(f"{language}")

    if evm_version is None:
        if language == "Solidity":
            evm_version = next(i[0] for i in EVM_SOLC_VERSIONS if solidity.get_version() >= i[1])
        else:
            evm_version = "petersburg"

    input_json: Dict = deepcopy(STANDARD_JSON)
    input_json["language"] = language
    input_json["settings"]["evmVersion"] = evm_version
    if language == "Solidity":
        input_json["settings"]["optimizer"] = {"enabled": optimize, "runs": runs if optimize else 0}
    input_json["sources"] = dict(
        (k, {"content": sources.minify(v, language)[0] if minify else v})
        for k, v in contract_sources.items()
    )
    return input_json


def compile_from_input_json(
    input_json: Dict, silent: bool = True, allow_paths: Optional[str] = None
) -> Dict:

    """
    Compiles contracts from a standard input json.

    Args:
        input_json: solc input json
        silent: verbose reporting
        allow_paths: compiler allowed filesystem import path

    Returns: standard compiler output json
    """

    if input_json["language"] == "Vyper":
        return vyper.compile_from_input_json(input_json, silent, allow_paths)
    if input_json["language"] == "Solidity":
        return solidity.compile_from_input_json(input_json, silent, allow_paths)
    raise UnsupportedLanguage(f"{input_json['language']}")


def generate_build_json(
    input_json: Dict, output_json: Dict, compiler_data: Optional[Dict] = None, silent: bool = True
) -> Dict:
    """Formats standard compiler output to the brownie build json.

    Args:
        input_json: solc input json used to compile
        output_json: output json returned by compiler
        compiler_data: additonal data to include under 'compiler' in build json
        silent: verbose reporting

    Returns: build json dict"""

    if input_json["language"] not in ("Solidity", "Vyper"):
        raise UnsupportedLanguage(f"{input_json['language']}")

    if not silent:
        print("Generating build data...")

    if compiler_data is None:
        compiler_data = {}
    compiler_data["evm_version"] = input_json["settings"]["evmVersion"]
    minified = compiler_data.get("minify_source", False)
    build_json = {}
    path_list = list(input_json["sources"])

    if input_json["language"] == "Solidity":
        compiler_data.update(
            {
                "optimize": input_json["settings"]["optimizer"]["enabled"],
                "runs": input_json["settings"]["optimizer"]["runs"],
            }
        )
        source_nodes, statement_nodes, branch_nodes = solidity._get_nodes(output_json)

    for path_str, contract_name in [
        (k, v) for k in path_list for v in output_json["contracts"].get(k, {})
    ]:

        if not silent:
            print(f" - {contract_name}...")

        abi = output_json["contracts"][path_str][contract_name]["abi"]
        output_evm = output_json["contracts"][path_str][contract_name]["evm"]
        hash_ = sources.get_hash(
            input_json["sources"][path_str]["content"],
            contract_name,
            minified,
            input_json["language"],
        )

        if input_json["language"] == "Solidity":
            contract_node = next(
                i[contract_name] for i in source_nodes if i.absolutePath == path_str
            )
            build_json[contract_name] = solidity._get_unique_build_json(
                output_evm,
                contract_node,
                statement_nodes,
                branch_nodes,
                next((True for i in abi if i["type"] == "fallback"), False),
            )

        else:
            if contract_name == "<stdin>":
                contract_name = "Vyper"
            build_json[contract_name] = vyper._get_unique_build_json(
                output_evm,
                path_str,
                contract_name,
                output_json["sources"][path_str]["ast"],
                (0, len(input_json["sources"][path_str]["content"])),
            )

        build_json[contract_name].update(
            {
                "abi": abi,
                "ast": output_json["sources"][path_str]["ast"],
                "compiler": compiler_data,
                "contractName": contract_name,
                "deployedBytecode": output_evm["deployedBytecode"]["object"],
                "deployedSourceMap": output_evm["deployedBytecode"]["sourceMap"],
                "language": input_json["language"],
                "opcodes": output_evm["deployedBytecode"]["opcodes"],
                "sha1": hash_,
                "source": input_json["sources"][path_str]["content"],
                "sourceMap": output_evm["bytecode"].get("sourceMap", ""),
                "sourcePath": path_str,
            }
        )

    if not silent:
        print("")

    return build_json
